from __future__ import annotations

from iotbx.detectors.hamamatsu import HamamatsuImage

from dxtbx.format.FormatSMVHamamatsu import FormatSMVHamamatsu


class FormatSMVHamamatsuSPring8BL32XU(FormatSMVHamamatsu):
    @staticmethod
    def understand(image_file):
        size, header = FormatSMVHamamatsu.get_smv_header(image_file)

        wanted_header_items = ["DETECTOR_NAME"]

        for header_item in wanted_header_items:
            if header_item not in header:
                return 0

        return header["DETECTOR_NAME"] == "Hamamatsu C10158DK"

    def _start(self):
        FormatSMVHamamatsu._start(self)

    def detectorbase_start(self):
        self.detectorbase = HamamatsuImage(self._image_file)
        self.detectorbase.open_file = self.open_file
        self.detectorbase.readHeader()

    def _goniometer(self):
        """Return a model for a simple single-axis reversed direction goniometer."""

        return self._goniometer_factory.single_axis_reverse()



"""Format classes to specifically recognise images from an electron detector
with a 2x2 array of Timepix modules, converted to SMV in various ways."""

from __future__ import annotations

import calendar
import sys
import time

from iotbx.detectors import SMVImage
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.FormatSMV import FormatSMV
from dxtbx.masking import mask_untrusted_rectangle
from dxtbx.model.beam import Probe
from dxtbx.model.detector import Detector


class FormatSMVTimePix_SU(FormatSMV):
    """Base format class to specifically recognise images from a Timepix-based
    detector, installed on a JEOL-2100 electron microscope at Stockholm
    University, which have been preprocessed by Wei Wan's RED software,
    to produce SMV files. See https://doi.org/10.1107/S0021889813027714 for RED.
    The detector itself has a 2x2 array of Timepix modules. Derived classes will
    either treat these separately as a multi-panel model, or as a single panel
    model."""

    @staticmethod
    def understand(image_file):
        size, header = FormatSMV.get_smv_header(image_file)

        # only recognise TimePix_SU
        if header.get("BEAMLINE", "").upper() != "TIMEPIX_SU":
            return False

        # check the header contains the things we're going to use
        wanted_header_items = [
            "BEAM_CENTER_X",
            "BEAM_CENTER_Y",
            "DISTANCE",
            "WAVELENGTH",
            "PIXEL_SIZE",
            "OSC_START",
            "OSC_RANGE",
            "PHI",
            "SIZE1",
            "SIZE2",
            "BYTE_ORDER",
            "DETECTOR_SN",
        ]
        if any(item not in header for item in wanted_header_items):
            return False

        return True

    def detectorbase_start(self):
        if not hasattr(self, "detectorbase") or self.detectorbase is None:
            self.detectorbase = SMVImage(self._image_file)
            self.detectorbase.open_file = self.open_file
            self.detectorbase.readHeader()

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. For this beamline
        this should be close to the provided values and neither aligned with the
        detector slow or fast axes."""

        return self._goniometer_factory.known_axis((-0.755, -0.656, 0.0))

    def _beam(self):
        """Return an unpolarized beam model."""

        wavelength = float(self._header_dictionary["WAVELENGTH"])

        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=Probe.electron,
        )

    def _scan(self):
        """Return the scan information for this image."""
        exposure_time = float(self._header_dictionary["TIME"])
        epoch = None

        # PST, PDT timezones not recognised by default...

        epoch = 0
        try:
            date_str = self._header_dictionary["DATE"]
            date_str = date_str.replace("PST", "").replace("PDT", "")
        except KeyError:
            date_str = ""
        for format_string in ["%a %b %d %H:%M:%S %Y", "%a %b %d %H:%M:%S %Z %Y"]:
            try:
                epoch = calendar.timegm(time.strptime(date_str, format_string))
                break
            except ValueError:
                pass

        # assert(epoch)
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = float(self._header_dictionary["OSC_RANGE"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )


class FormatSMVTimePix_SU_512x512(FormatSMVTimePix_SU):
    """Implementation of a format class to specifically recognise images from a
    Timepix-based detector, installed on a JEOL-2100 electron microscope at
    Stockholm University, which have been preprocessed by Wei Wan's RED software,
    to produce SMV files. See https://doi.org/10.1107/S0021889813027714 for RED.
    The detector itself has a 2x2 array of Timepix modules."""

    @staticmethod
    def understand(image_file):
        size, header = FormatSMVTimePix_SU.get_smv_header(image_file)

        # check the pixel size is 55 microns
        if float(header["PIXEL_SIZE"]) != 0.055:
            return False

        # check there are 512*512 pixels
        if header["SIZE1"] != "512":
            return False
        if header["SIZE2"] != "512":
            return False

        return True

    def _detector(self):
        """4 panel detector, 55 micron pixels except for pixels at the outer
        edge of each chip, which are 165 microns wide."""
        # expect 55 mu pixels here, but make it general anyway
        pixel_size = tuple([float(self._header_dictionary["PIXEL_SIZE"])] * 2)
        image_size = (
            int(self._header_dictionary["SIZE1"]),
            int(self._header_dictionary["SIZE2"]),
        )
        panel_size = tuple([int(e / 2) for e in image_size])

        # outer pixels have three times the width
        # panel_size_mm = (
        #    pixel_size[0] * 3 + (panel_size[0] - 2) * pixel_size[0],
        #    pixel_size[1] * 3 + (panel_size[1] - 2) * pixel_size[1],
        # )
        trusted_range = (0, 65535)
        thickness = 0.3  # assume 300 mu thick

        # Initialise detector frame
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, -1.0, 0.0))
        beam_centre = (
            float(self._header_dictionary["BEAM_CENTER_X"]),
            float(self._header_dictionary["BEAM_CENTER_Y"]),
        )

        bx_px, by_px = beam_centre

        def px_to_mm(px, px_size_1d, panel_size_1d):
            # the beam centre is in pixels. We want to convert to mm, taking the
            # different size of outer pixels into account. Use this local function
            # to do that
            mm = 0
            if px > 1:  # add first outer pixel
                mm += px_size_1d * 3
            else:  # or fraction of first outer pixel
                mm += px * px_size_1d * 3
                return mm

            if px > panel_size_1d - 1:  # add full panel of inner pixels
                mm += (panel_size_1d - 2) * px_size_1d
            else:  # or fraction of inner pixels
                mm += (px - 1) * px_size_1d
                return mm

            if px > panel_size_1d:  # add second outer pixel
                mm += px_size_1d * 3
            else:  # or fraction of second outer pixel
                mm += (px - (panel_size_1d - 1)) * px_size_1d * 3
                return mm

            if px > panel_size_1d + 1:  # add first outer pixel of second panel
                mm += px_size_1d * 3
            else:  # or fraction of first outer pixel of second panel
                mm += (px - panel_size_1d) * px_size_1d * 3
                return mm

            if px > (2 * panel_size_1d - 1):  # add second full panel of inner pixels
                mm += (panel_size_1d - 2) * px_size_1d
                # plus remaining fraction of the second outer pixel
                mm += (px - (2 * panel_size_1d - 1)) * px_size_1d * 3
            else:  # or fraction of inner pixels of the second panel
                mm += (px - panel_size_1d - 1) * px_size_1d
            return mm

        bx_mm = px_to_mm(bx_px, pixel_size[0], panel_size[0])
        by_mm = px_to_mm(by_px, pixel_size[1], panel_size[1])

        # the beam centre is defined from the origin along fast, slow. To determine
        # the lab frame origin we place the beam centre down the -z axis
        dist = float(self._header_dictionary["DISTANCE"])
        cntr = matrix.col((0.0, 0.0, -1 * dist))
        orig = cntr - bx_mm * fast - by_mm * slow

        d = Detector()

        root = d.hierarchy()
        root.set_local_frame(fast.elems, slow.elems, orig.elems)

        self.coords = {}
        panel_idx = 0

        # set panel extent in pixel numbers and x, y mm shifts. Note that the
        # outer pixels are 0.165 mm in size. These are excluded from the panel
        # extents.
        pnl_data = []
        pnl_data.append(
            {
                "xmin": 1,
                "ymin": 1,
                "xmax": 255,
                "ymax": 255,
                "xmin_mm": 1 * 0.165,
                "ymin_mm": 1 * 0.165,
            }
        )
        pnl_data.append(
            {
                "xmin": 257,
                "ymin": 1,
                "xmax": 511,
                "ymax": 255,
                "xmin_mm": 3 * 0.165 + (511 - 257) * pixel_size[0],
                "ymin_mm": 1 * 0.165,
            }
        )
        pnl_data.append(
            {
                "xmin": 1,
                "ymin": 257,
                "xmax": 255,
                "ymax": 511,
                "xmin_mm": 1 * 0.165,
                "ymin_mm": 3 * 0.165 + (511 - 257) * pixel_size[1],
            }
        )
        pnl_data.append(
            {
                "xmin": 257,
                "ymin": 257,
                "xmax": 511,
                "ymax": 511,
                "xmin_mm": 3 * 0.165 + (511 - 257) * pixel_size[0],
                "ymin_mm": 3 * 0.165 + (511 - 257) * pixel_size[1],
            }
        )

        # redefine fast, slow for the local frame
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, 1.0, 0.0))

        for ipanel, pd in enumerate(pnl_data):
            xmin = pd["xmin"]
            xmax = pd["xmax"]
            ymin = pd["ymin"]
            ymax = pd["ymax"]
            xmin_mm = pd["xmin_mm"]
            ymin_mm = pd["ymin_mm"]

            origin_panel = fast * xmin_mm + slow * ymin_mm

            panel_name = "Panel%d" % panel_idx
            panel_idx += 1

            p = d.add_panel()
            p.set_type("SENSOR_PAD")
            p.set_name(panel_name)
            p.set_raw_image_offset((xmin, ymin))
            p.set_image_size((xmax - xmin, ymax - ymin))
            p.set_trusted_range(trusted_range)
            p.set_pixel_size((pixel_size[0], pixel_size[1]))
            p.set_thickness(thickness)
            p.set_material("Si")
            # p.set_mu(mu)
            # p.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
            p.set_local_frame(fast.elems, slow.elems, origin_panel.elems)
            p.set_raw_image_offset((xmin, ymin))
            self.coords[panel_name] = (xmin, ymin, xmax, ymax)

        return d

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        raw_data = self._get_endianic_raw_data(size=(512, 512))

        # split into separate panels
        self._raw_data = []
        d = self.get_detector()
        for panel in d:
            xmin, ymin, xmax, ymax = self.coords[panel.get_name()]
            self._raw_data.append(raw_data[ymin:ymax, xmin:xmax])

        return tuple(self._raw_data)


class FormatSMVTimePix_SU_516x516(FormatSMVTimePix_SU):
    """A class for reading SMV format images for the Timepix-based electron
    detector where the wider edge pixels have been split into three normal-sized
    pixels. The whole 516*516 detector can then be described as a single panel.
    We choose to mask out the edge pixels, as they do not correctly represent
    reflection profiles."""

    @staticmethod
    def understand(image_file):
        size, header = FormatSMVTimePix_SU.get_smv_header(image_file)

        # check there are 516*516 pixels
        if header["SIZE1"] != "516":
            return False
        if header["SIZE2"] != "516":
            return False

        return True

    def get_static_mask(self):
        """Return the static mask that excludes the central cross of pixels."""

        mask = flex.bool(flex.grid((516, 516)), True)
        mask_untrusted_rectangle(mask, 0, 516, 255, 261)
        mask_untrusted_rectangle(mask, 255, 261, 0, 516)

        return (mask,)

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. For this beamline
        this should be close to the provided values and neither aligned with the
        detector slow or fast axes."""

        # import numpy as np
        # angle = np.radians(-40) #-0.672  # radians
        # axis = np.cos(angle), -np.cos(angle+np.pi/2), 0

        # axis = (0.755, -0.656, 0.0)
        axis = (0.7826, -0.6226, 0.0)
        return self._goniometer_factory.known_axis(axis)

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])  # mm

        beam_x = float(self._header_dictionary["BEAM_CENTER_X"])  # px
        beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])  # px

        # wavelength = float(self._header_dictionary['WAVELENGTH'])  # Angstrom

        pixelsize = float(self._header_dictionary["PIXEL_SIZE"])  # mm

        thickness = 0.3  # mm
        material = "Si"

        nx = int(self._header_dictionary["SIZE1"])  # number of pixels
        ny = int(self._header_dictionary["SIZE2"])  # number of pixels

        underload, overload = (-1, 65535)

        detector = self._detector_factory.simple(
            sensor="PAD",
            distance=distance,
            beam_centre=(beam_x * pixelsize, beam_y * pixelsize),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(pixelsize, pixelsize),
            image_size=(nx, ny),
            trusted_range=(underload, overload),
        )

        detector[0].set_thickness(thickness)
        detector[0].set_material(material)
        detector[0].mask = []  # can we mask the cross here?

        return detector

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        nx = int(self._header_dictionary["SIZE1"])  # number of pixels
        ny = int(self._header_dictionary["SIZE2"])  # number of pixels
        raw_data = self._get_endianic_raw_data(size=(nx, ny))

        self._raw_data = raw_data

        return self._raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print("FormatSMVTimePix_SU:", FormatSMVTimePix_SU.understand(arg))
        print(
            "FormatSMVTimePix_SU_512x512:", FormatSMVTimePix_SU_512x512.understand(arg)
        )
        print(
            "FormatSMVTimePix_SU_516x516:", FormatSMVTimePix_SU_516x516.understand(arg)
        )



"""An implementation of the CBF image reader for Eiger images"""

from __future__ import annotations

import binascii
import os
import sys

from cctbx.eltbx import attenuation_coefficient
from iotbx.detectors.eiger_minicbf import EigerCBFImage

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFMini import FormatCBFMini
from dxtbx.format.FormatCBFMiniPilatusHelpers import get_pilatus_timestamp
from dxtbx.format.FormatPilatusHelpers import determine_eiger_mask
from dxtbx.format.FormatPilatusHelpers import get_vendortype_eiger as gv
from dxtbx.model import ParallaxCorrectedPxMmStrategy

if "DXTBX_OVERLOAD_SCALE" in os.environ:
    dxtbx_overload_scale = float(os.environ["DXTBX_OVERLOAD_SCALE"])
else:
    dxtbx_overload_scale = 1


class FormatCBFMiniEiger(FormatCBFMini):
    """A class for reading mini CBF format Eiger images, and correctly
    constructing a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Eiger mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMini.get_cbf_header(image_file)

        for record in header.split("\n"):
            if "# detector" in record.lower() and "eiger" in record.lower():
                return True

        return False

    def _detector(self):
        distance = float(self._cif_header_dictionary["Detector_distance"].split()[0])

        beam_xy = (
            self._cif_header_dictionary["Beam_xy"]
            .replace("(", "")
            .replace(")", "")
            .replace(",", "")
            .split()[:2]
        )

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        beam_x, beam_y = map(float, beam_xy)

        pixel_xy = (
            self._cif_header_dictionary["Pixel_size"]
            .replace("m", "")
            .replace("x", "")
            .split()
        )

        pixel_x, pixel_y = map(float, pixel_xy)

        if "Silicon" in self._cif_header_dictionary:
            thickness = (
                float(self._cif_header_dictionary["Silicon"].split()[2]) * 1000.0
            )
            material = "Si"
        elif "CdTe" in self._cif_header_dictionary:
            thickness = float(self._cif_header_dictionary["CdTe"].split()[2]) * 1000.0
            material = "CdTe"
        else:
            thickness = 0.450
            material = "Si"

        nx = int(self._cif_header_dictionary["X-Binary-Size-Fastest-Dimension"])
        ny = int(self._cif_header_dictionary["X-Binary-Size-Second-Dimension"])

        if "Count_cutoff" in self._cif_header_dictionary:
            overload = int(self._cif_header_dictionary["Count_cutoff"].split()[0])
        else:
            # missing from data transformed with GPhL converter - dials#376
            overload = 100000000
        minimum_trusted_value = 0

        try:
            identifier = self._cif_header_dictionary["Detector"].encode()
        except KeyError:
            identifier = "Unknown Eiger"

        table = attenuation_coefficient.get_table(material)
        mu = table.mu_at_angstrom(wavelength) / 10.0
        t0 = thickness

        detector = self._detector_factory.simple(
            "PAD",
            distance * 1000.0,
            (beam_x * pixel_x * 1000.0, beam_y * pixel_y * 1000.0),
            "+x",
            "-y",
            (1000 * pixel_x, 1000 * pixel_y),
            (nx, ny),
            (minimum_trusted_value, overload),
            [],
            px_mm=ParallaxCorrectedPxMmStrategy(mu, t0),
            mu=mu,
        )

        for f0, f1, s0, s1 in determine_eiger_mask(detector):
            detector[0].add_mask(f0 - 1, s0 - 1, f1, s1)

        for panel in detector:
            panel.set_thickness(thickness)
            panel.set_material(material)
            panel.set_identifier(identifier)
            panel.set_mu(mu)

        return detector

    def _beam(self):
        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        beam = self._beam_factory.simple(wavelength)

        try:
            flux = float(self._cif_header_dictionary["Flux"].split()[0])
            beam.set_flux(flux)
        except KeyError:
            pass

        try:
            transmission = float(self._cif_header_dictionary["Transmission"].split()[0])
            beam.set_transmission(transmission)
        except KeyError:
            pass

        return beam

    def _scan(self):
        exposure_time = float(self._cif_header_dictionary["Exposure_period"].split()[0])
        osc_start = float(self._cif_header_dictionary["Start_angle"].split()[0])
        osc_range = float(self._cif_header_dictionary["Angle_increment"].split()[0])

        if "timestamp" in self._cif_header_dictionary:
            timestamp = get_pilatus_timestamp(self._cif_header_dictionary["timestamp"])
        else:
            timestamp = 0.0

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, timestamp
        )

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values

    def get_raw_data(self):
        if self._raw_data is None:
            data = self._read_cbf_image()
            self._raw_data = data

        return self._raw_data

    def detectorbase_start(self):
        self.detectorbase = EigerCBFImage(self._image_file)
        self.detectorbase.readHeader()

    def get_vendortype(self):
        return gv(self.get_detector())


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniEiger.understand(arg))



from __future__ import annotations

import weakref

import h5py
import nxmx

import scitbx.array_family.flex as flex

import dxtbx.nexus
from dxtbx.format.FormatNexus import FormatNexus


class _MaskCache:
    """A singleton to hold unique static_mask objects to avoid duplications"""

    def __init__(self):
        self.local_mask_cache = weakref.WeakValueDictionary()

    def _mask_hasher(self, mask: flex.bool) -> int:
        return hash(mask.as_numpy_array().tobytes())

    def store_unique_and_get(
        self, mask_tuple: tuple[flex.bool, ...] | None
    ) -> tuple[flex.bool, ...] | None:
        if mask_tuple is None:
            return None
        output = []
        for mask in mask_tuple:
            mask_hash = self._mask_hasher(mask)
            mask = self.local_mask_cache.setdefault(mask_hash, mask)
            output.append(mask)
        return tuple(output)


mask_cache = _MaskCache()


def detector_between_sample_and_source(detector, beam):
    """Check if the detector is perpendicular to beam and
    upstream of the sample."""

    if len(detector) != 1:
        return False
    p = detector[0]

    n = p.get_normal()
    o = p.get_origin()
    x = beam.get_sample_to_source_direction()

    dot = sum(_n * _x for _n, _x in zip(n, x))
    doto = sum(_o * _x for _o, _x in zip(o, x))

    if abs(dot) > 0.99 and doto > 0:
        return True

    return False


def inverted_distance_detector(detector):
    """Move the detector origin to x, y, -z"""

    origin = detector[0].get_origin()
    origin = origin[0], origin[1], -origin[2]
    fast = detector[0].get_fast_axis()
    slow = detector[0].get_slow_axis()
    detector[0].set_frame(fast, slow, origin)
    return detector


class FormatNXmx(FormatNexus):
    """Read NXmx-flavour NeXus-format HDF5 data."""

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file) as handle:
            return bool(
                [
                    entry
                    for entry in nxmx.find_class(handle, "NXentry")
                    if "definition" in entry
                    and nxmx.h5str(entry["definition"][()]) == "NXmx"
                ]
            )

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""
        super().__init__(image_file, **kwargs)

    def _start(self):
        self._static_mask = None

        self._cached_file_handle = h5py.File(self._image_file, swmr=True)
        nxmx_obj = self._get_nxmx(self._cached_file_handle)
        nxentry = nxmx_obj.entries[0]
        nxsample = nxentry.samples[0]
        nxinstrument = nxentry.instruments[0]
        nxdetector = nxinstrument.detectors[0]
        nxbeam = nxinstrument.beams[0]
        nxdata = nxmx_obj.entries[0].data[0]
        self._goniometer_model = dxtbx.nexus.get_dxtbx_goniometer(nxsample)
        self._beam_factory = dxtbx.nexus.CachedWavelengthBeamFactory(nxbeam)
        wavelength = self._beam_factory.make_beam(index=0).get_wavelength()
        self._detector_model = dxtbx.nexus.get_dxtbx_detector(
            nxdetector, wavelength, nxdata
        )

        # if the detector is between the sample and the source, and perpendicular
        # to the beam, then invert the distance vector, as this is probably wrong
        beam = self._beam()
        if detector_between_sample_and_source(self._detector_model, beam):
            self._detector_model = inverted_distance_detector(self._detector_model)

        self._scan_model = dxtbx.nexus.get_dxtbx_scan(nxsample, nxdetector)
        self._static_mask = mask_cache.store_unique_and_get(
            dxtbx.nexus.get_static_mask(nxdetector)
        )
        self._bit_depth_readout = nxdetector.bit_depth_readout

        if self._scan_model:
            self._num_images = len(self._scan_model)
        else:
            if nxdata.signal:
                data = nxdata[nxdata.signal]
            else:
                data = list(nxdata.values())[0]
            self._num_images, *_ = data.shape

        # handle reversed rotation by inverting both the rotation angles
        # and the rotation axis
        if self._scan_model:
            scan = self._scan_model
            oscillation = scan.get_oscillation()

            if oscillation[1] < 0:
                # invert the scan parameters so that the attitude of roation is positive
                self._scan_model.set_oscillation((-oscillation[0], -oscillation[1]))

                # invert the axis direction so that the universe is aligned
                axes = self._goniometer_model.get_axes()
                axis = self._goniometer_model.get_scan_axis()
                axes[axis] = -1 * axes[axis][0], -1 * axes[axis][1], -1 * axes[axis][2]
                self._goniometer_model.set_axes(axes)

    def _get_nxmx(self, fh: h5py.File):
        return nxmx.NXmx(fh)

    def _beam(self, index: int | None = None) -> dxtbx.model.Beam:
        return self._beam_factory.make_beam(index=index or 0)

    def get_spectrum(self, index=None):
        return self._beam_factory.make_spectrum(index=index or 0)

    def get_num_images(self) -> int:
        return self._num_images

    def get_static_mask(self, index=None, goniometer=None):
        return self._static_mask

    def get_raw_data(self, index):
        nxmx_obj = self._get_nxmx(self._cached_file_handle)
        nxdata = nxmx_obj.entries[0].data[0]
        nxdetector = nxmx_obj.entries[0].instruments[0].detectors[0]
        raw_data = dxtbx.nexus.get_raw_data(
            nxdata, nxdetector, index, bit_depth=self._bit_depth_readout
        )
        if self._bit_depth_readout:
            # if 32 bit then it is a signed int, I think if 8, 16 then it is
            # unsigned with the highest two values assigned as masking values
            if self._bit_depth_readout == 32:
                top = 2**31
            else:
                top = 2**self._bit_depth_readout
            for data in raw_data:
                d1d = data.as_1d()
                d1d.set_selected(d1d == top - 1, -1)
                d1d.set_selected(d1d == top - 2, -2)
        return raw_data



# Following taken from:
#
# http://www.rigaku.com/downloads/software/readimage.html
#
# Header Structure:
#
# First 256 bytes
#
# 52 characters: name (10) version (10) xname (20) system (12)
# 6 floats: a, b, c, alpha, beta, gamma
# 12 characters: spacegroup
# float: mosaic
# 80 characters: notes
# 84 characters: padding
#
# Next 256 bytes:
#
# 36 characters: date (12) username (20) target (4)
# float wavelength
# 20 character: typo of mono
# float: mono 2theta (degrees)
# 24 characters: collimator settings (20) filter type (4)
# 3 floats: distance, voltage, current
# 92 characters: focus stuff (12) xray chunder (80)
# long int: ipshape - 0 flat (good) 1 cylinder (bad)
# float weiss (1?)
# 56 characters: padding
#
# Next 256 bytes:
#
# 8 bytes (xtal axes)
# 3 float - phi0, phistart, phiend
# long int frame no
# float exposure time (minutes!)
# 6 floats - beam x, beam z omega chi two theta mu
# 204 characters - padding
#
# Next 256 bytes:
#
# 2 long ints - nx, nz
# 2 float - pixel size
# 4 long, 3 float (not really useful)
# 20 char host (10) ip type (10)
# 3 long - scan order horizontal, vertical, back / front:
#
#  long  dr_x;    /* horizontal scanning code: 0=left->right, 1=>right->left */
#  long  dr_z;    /* vertical scanning code: 0=down->up, 1=up->down */
#  long  drxz;    /* front/back scanning code: 0=front, 1=back */
#
# 2 useless floats
# 2 long - a magic number to show validity of this section, and number of axes
# 15 floats - up to 5 goniometer axis vectors
# 5 floats - up to 5 start angles
# 5 floats - up to 5 end angles
# 5 floats - up to 5 offset values
# long - which axis is scan axis
# 40 characters - gonio axis names, space or comma separated
#
# Then some more chunder follows - however I don't think it contains anything
# useful. So need to read first 1K of the image header.


from __future__ import annotations

import calendar
import datetime
import io
import math
import struct

from iotbx.detectors import ReadRAXIS
from iotbx.detectors.raxis import RAXISImage

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format

header_struct = [
    ("Device", 10, "s"),
    ("Version", 10, "s"),
    ("Crystal", 20, "s"),
    ("CrystalSystem", 12, "s"),
    (None, 24),
    ("SpaceGroup", 12, "s"),
    ("mosaic1", 4, "!f"),
    ("memo", 80, "s"),
    ("reserve1", 84, "s"),
    ("date", 12, "s"),
    ("operatorname", 20, "s"),
    ("target", 4, "s"),
    ("wavelength", 4, "!f"),
    ("monotype", 20, "s"),
    ("mono2theta", 4, "!f"),
    ("collimator", 20, "s"),
    ("filter", 4, "s"),
    ("distance", 4, "!f"),
    ("Kv", 4, "!f"),
    ("mA", 4, "!f"),
    ("focus", 12, "s"),
    ("Xmemo", 80, "s"),
    ("cyl", 4, "!i"),
    (None, 60),
    ("Spindle", 4, "s"),  # Crystal mount axis closest to spindle axis
    ("Xray_axis", 4, "s"),  # Crystal mount axis closest to beam axis
    ("phidatum", 4, "!f"),
    ("phistart", 4, "!f"),
    ("phiend", 4, "!f"),
    ("noscillations", 4, "!i"),
    ("minutes", 4, "!f"),  # Exposure time in minutes?
    ("beampixels_x", 4, "!f"),
    ("beampixels_y", 4, "!f"),  # Direct beam position in pixels
    ("omega", 4, "!f"),
    ("chi", 4, "!f"),
    ("twotheta", 4, "!f"),
    ("Mu", 4, "!f"),  # Spindle inclination angle?
    ("ScanTemplate", 204, "s"),  # This space is now used for storing the scan
    # templates information
    ("nFast", 4, "!i"),
    ("nSlow", 4, "!i"),  # Number of fast, slow pixels
    ("sizeFast", 4, "!f"),
    ("sizeSlow", 4, "!f"),  # Size of fast, slow direction in mm
    ("record_length", 4, "!i"),  # Record length in bytes
    ("number_records", 4, "!i"),  # number of records
    ("Read_start", 4, "!i"),  # For partial reads, 1st read line
    ("IP_num", 4, "!i"),  # Which imaging plate 1, 2 ?
    ("Ratio", 4, "!f"),  # Output ratio for high value pixels
    ("Fading_start", 4, "!f"),  # Fading time to start of read
    ("Fading_end", 4, "!f"),  # Fading time to end of read
    ("computer", 10, "s"),  # Type of computer "IRIS", "VAX", "SUN", etc
    ("plate_type", 10, "s"),  # Type of IP
    ("Dr", 4, "!i"),
    ("Dx", 4, "!i"),
    ("Dz", 4, "!i"),  # IP scanning codes??
    ("PixShiftOdd", 4, "!f"),  # Pixel shift to odd lines
    ("IntRatioOdd", 4, "!f"),  # Intensity ratio to odd lines
    ("MagicNum", 4, "!i"),  # Magic number to indicate next values are legit
    ("NumGonAxes", 4, "!i"),  # Number of goniometer axes
    ("a5x3fGonVecs", 60, "!fffffffffffffff"),  # Goniometer axis vectors
    ("a5fGonStart", 20, "!fffff"),  # Start angles for each of 5 axes
    ("a5fGonEnd", 20, "!fffff"),  # End angles for each of 5 axes
    ("a5fGonOffset", 20, "!fffff"),  # Offset values for each of 5 axes
    ("ScanAxisNum", 4, "!i"),  # Which axis is the scan axis?
    ("AxesNames", 40, "s"),  # Names of the axes (space or comma separated?)'''
]


class RAXISHelper:
    def __init__(self, image_file, **kwargs):
        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)

        super().__init__(image_file, **kwargs)

    def _start(self):
        with Format.open_file(self._image_file) as fh:
            self._header_bytes = fh.read(1024)
        self.head = {}
        stream = io.BytesIO(self._header_bytes)
        for item in header_struct:
            if item[0] is None:
                stream.read(item[1])
            elif item[2] == "s":
                self.head[item[0]] = stream.read(item[1])[0 : item[1]]
            elif len(item[2]) > 2:
                rawdata = stream.read(item[1])
                assert len(rawdata) == struct.calcsize(item[2])
                self.head[item[0]] = struct.unpack(item[2], rawdata)
            else:
                rawdata = stream.read(item[1])
                assert len(rawdata) == struct.calcsize(item[2])
                self.head[item[0]] = struct.unpack(item[2], rawdata)[0]
        if self.head["computer"].strip() in (b"SGI", b"IRIS"):
            self._f = ">f"
            self._i = ">i"
        else:
            self._f = "<f"
            self._i = "<i"

    def detectorbase_start(self):
        self.detectorbase = RAXISImage(self._image_file)
        self.detectorbase.readHeader()

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a flex array."""
        self.detectorbase_start()

        dim0 = self.head["nFast"]  # number of fast pixels
        to_read = self.head["record_length"]
        read_lines = self.head["number_records"]

        with Format.open_file(self._image_file) as fh:
            fh.seek(to_read)
            raw_data = fh.read(to_read * read_lines)

        # For a normal image, there should be no padding per line
        # Each line might be padded, so figure this out
        bytes_per_line = dim0 * 2
        if bytes_per_line < to_read:
            # Remove all padding bytes
            raw_data = b"".join(
                raw_data[record * to_read : record * to_read + bytes_per_line]
                for record in range(read_lines)
            )

        raw_data = ReadRAXIS(
            raw_data,
            self.head["record_length"] // self.head["nFast"],
            self.head["nSlow"] * self.detectorbase.bin,
            self.head["nFast"] * self.detectorbase.bin,
            self.detectorbase.endian_swap_required(),
        )

        return raw_data

    def _detector_helper(self):
        """Returns image header values as a dictionary."""

        locations = {
            "det_h": (self._i, 832),
            "det_v": (self._i, 836),
            "det_f": (self._i, 840),
            "nx": (self._i, 768),
            "ny": (self._i, 772),
            "dx": (self._f, 776),
            "dy": (self._f, 780),
            "distance": (self._f, 344),
            "two_theta": (self._f, 556),
            "beam_x": (self._f, 540),
            "beam_y": (self._f, 544),
        }
        values = {
            name: struct.unpack(
                location[0], self._header_bytes[location[1] : location[1] + 4]
            )[0]
            for name, location in locations.items()
        }

        assert values["det_h"] == 0
        assert values["det_v"] == 0
        assert values["det_f"] == 0

        values["beam"] = (
            values["beam_x"] * values["dx"],
            values["beam_y"] * values["dy"],
        )
        return values


class FormatRAXIS(RAXISHelper, Format):
    """A class to support the RAXIS detector format from Rigaku."""

    @staticmethod
    def understand(image_file):
        """See if this looks like an RAXIS format image - clue is first
        5 letters of file should be RAXIS."""

        with Format.open_file(image_file) as fh:
            return fh.read(5) == b"RAXIS"

    def _goniometer(self):
        """Return a model for the goniometer from the values stored in the
        header. Assumes same reference frame as is used for the Saturn
        instrument."""

        # OK for the moment I am not completely clear on how omega, chi and
        # phi are defined so for the moment assert (i) that only one rotation
        # axis has a non-zero offset (ii) that this is the scan axis and
        # (iii) that this is aligned with (1, 0, 0) in the file... N.B.
        # with these scanners it is (I believe) always the case that the
        # (1, 0, 0) direction points downwards from the sample to the
        # goniometer.

        i = self._i
        f = self._f
        header = self._header_bytes

        # check magic number to see if this section is valid
        if struct.unpack(i, header[852:856]) == 1:
            n_axes = struct.unpack(i, header[856:860])[0]
        else:
            n_axes = 0
        scan_axis = struct.unpack(i, header[980:984])[0]

        for j in range(n_axes):
            axis_x = struct.unpack(f, header[860 + j * 12 : 864 + j * 12])[0]
            axis_y = struct.unpack(f, header[864 + j * 12 : 868 + j * 12])[0]
            axis_z = struct.unpack(f, header[868 + j * 12 : 872 + j * 12])[0]

            axis_start = struct.unpack(f, header[920 + j * 4 : 924 + j * 4])[0]
            axis_end = struct.unpack(f, header[940 + j * 4 : 944 + j * 4])[0]
            # axis_offset = struct.unpack(f, header[960 + j * 4 : 964 + j * 4])[0]

            if j == scan_axis:
                assert math.fabs(axis_x - 1) < 0.001
                assert math.fabs(axis_y) < 0.001
                assert math.fabs(axis_z) < 0.001
            else:
                assert math.fabs(axis_start % 180.0) < 0.001
                assert math.fabs(axis_end % 180.0) < 0.001

        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for the detector as defined in the image header,
        with the additional knowledge about how things are arranged i.e. that
        the principle rotation axis vector points from the sample downwards."""

        # As above will have to make a bunch of assumptions about how the
        # detector is set up, namely that the rotation axis points downwards,
        # the fast axis points along and the slow axis points up. These will
        # (as well as I can understand) be asserted. Also assuming that the
        # two-theta axis is along (1, 0, 0) which is all that makes sense.

        values = self._detector_helper()

        return self._detector_factory.two_theta(
            "IMAGE_PLATE",
            values["distance"],
            values["beam"],
            "+y",
            "-x",
            "+x",
            values["two_theta"],
            (values["dx"], values["dy"]),
            (values["nx"], values["ny"]),
            (0, 1000000),
            [],
        )

    def _beam(self):
        """Return a simple model for the beam. This assumes a laboratory source
        which has an unpolarized beam."""

        wavelength = struct.unpack(self._f, self._header_bytes[292:296])[0]

        return self._beam_factory.complex(
            (0.0, 0.0, 1.0), 0.5, (0.0, 1.0, 0.0), wavelength
        )

    def _scan(self):
        """Return the scan information for this image."""
        i = self._i
        f = self._f
        header = self._header_bytes

        # We expect an invalid goniometer section, indicated by wrong magic number
        magic_no = struct.unpack(">i", header[852:856])

        exposure_time = struct.unpack(f, header[536:540])[0]

        y, m, d = map(int, header[256:268].strip().split(b"-"))

        epoch = calendar.timegm(datetime.datetime(y, m, d, 0, 0, 0).timetuple())

        if magic_no == 1:
            s = struct.unpack(i, header[980:984])[0]
            osc_start = struct.unpack(f, header[920 + s * 4 : 924 + s * 4])[0]
            osc_end = struct.unpack(f, header[940 + s * 4 : 944 + s * 4])[0]

        else:
            osc_dat = struct.unpack(f, header[520:524])[0]
            osc_start = osc_dat + struct.unpack(f, header[524:528])[0]
            osc_end = osc_dat + struct.unpack(f, header[528:532])[0]

        osc_range = osc_end - osc_start

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )



"""
Dynamically generate the list of console_scripts dxtbx.format entry-points.
"""

from __future__ import annotations

import ast
import re
from pathlib import Path

from hatchling.metadata.plugin.interface import MetadataHookInterface


def get_entry_point(
    filename: Path, prefix: str, import_path: str
) -> list[tuple[str, str]]:
    """Returns any entry point strings for a given path.

    This looks for LIBTBX_SET_DISPATCHER_NAME, and a root function
    named 'run'. It can return multiple results for each file, if more
    than one dispatcher name is bound.

    Args:
        filename:
            The python file to parse. Will look for a run() function
            and any number of LIBTBX_SET_DISPATCHER_NAME.
        prefix: The prefix to output the entry point console script with
        import_path: The import path to get to the package the file is in

    Returns:
        A list of entry_point specifications
    """
    contents = filename.read_text()
    tree = ast.parse(contents)
    # Find root functions named "run"
    has_run = any(
        x
        for x in tree.body
        if (isinstance(x, ast.FunctionDef) and x.name == "run")
        or (isinstance(x, ast.ImportFrom) and "run" in [a.name for a in x.names])
    )
    if not has_run:
        return []
    # Find if we need an alternate name via LIBTBX_SET_DISPATCHER_NAME
    alternate_names = re.findall(
        r"^#\s*LIBTBX_SET_DISPATCHER_NAME\s+(.*)$", contents, re.M
    )
    if alternate_names:
        return [
            (name, f"{import_path}.{filename.stem}:run") for name in alternate_names
        ]

    return [(f"{prefix}.{filename.stem}", f"{import_path}.{filename.stem}:run")]


def enumerate_format_classes(path: Path) -> list[tuple[str, str]]:
    """Find all Format*.py files and contained Format classes in a path"""
    format_classes = []
    for filename in path.glob("Format*.py"):
        content = filename.read_bytes()
        try:
            parsetree = ast.parse(content)
        except SyntaxError:
            print(f"  *** Could not parse {filename.name}")
            continue
        for top_level_def in parsetree.body:
            if not isinstance(top_level_def, ast.ClassDef):
                continue
            base_names = [
                baseclass.id
                for baseclass in top_level_def.bases
                if isinstance(baseclass, ast.Name) and baseclass.id.startswith("Format")
            ]
            if base_names:
                classname = top_level_def.name
                format_classes.append(
                    (
                        f"{classname}:{','.join(base_names)}",
                        f"dxtbx.format.{filename.stem}:{classname}",
                    )
                )
    return format_classes


class CustomMetadataHook(MetadataHookInterface):
    def update(self, metadata):
        scripts = metadata.setdefault("scripts", {})
        package_path = Path(self.root) / "src" / "dxtbx"
        for file in package_path.joinpath("command_line").glob("*.py"):
            for name, symbol in get_entry_point(file, "dxtbx", "dxtbx.command_line"):
                if name not in scripts:
                    scripts[name] = symbol

        plugins = metadata.setdefault("entry-points", {})
        formats = plugins.setdefault("dxtbx.format", {})
        for name, symbol in sorted(enumerate_format_classes(package_path / "format")):
            if name not in formats:
                formats[name] = symbol



"""Byte offset implementation of fullCBF format, for use with CBF files using
byte offset compression which are _not_ made by dectris."""

from __future__ import annotations

import binascii
import os
from io import TextIOWrapper
from typing import IO

from scitbx.array_family import flex

from dxtbx.ext import uncompress
from dxtbx.format.FormatCBFFull import FormatCBFFull


def _read_cif_binary_format_section(file: IO[bytes]) -> str:
    """
    Read a CIF binary format section

    This is a MIME-variant.
    """
    header = ""
    reader = TextIOWrapper(file, encoding="ascii", errors="surrogateescape")
    for line in reader:
        if not line.strip():
            break
        header = header + line
    return header


class FormatCBFFullByteOffset(FormatCBFFull):
    """An image reading class for full CBF format images with byte-offset compression"""

    @staticmethod
    def understand(image_file: str | bytes | os.PathLike) -> bool:
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBFFull.get_cbf_header(image_file)

        # If this is a pilatus cbf file, then we use a different format
        for record in header.split("\n"):
            if "_array_data.header_convention" in record and "PILATUS" in record:
                return False

        with FormatCBFFull.open_file(image_file, "rb") as file:
            # If we've gotten a header, this ends at the MIME header
            file.seek(len(header))
            mime_divider = b"--CIF-BINARY-FORMAT-SECTION--\r\n"
            if not file.read(len(mime_divider)) == mime_divider:
                return False

            binary_header = _read_cif_binary_format_section(file)

        # If we've got this signature in the binary header, we can read
        if 'conversions="x-CBF_BYTE_OFFSET"' in binary_header:
            return True

        return False

    def _read_cbf_image(self) -> flex.int:
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        pixel_values = uncompress(
            packed=data[data_offset : data_offset + cbf_header["size"]],
            fast=cbf_header["fast"],
            slow=cbf_header["slow"],
        )

        return pixel_values

    def get_raw_data(self) -> flex.int:
        if self._raw_data is None:
            data = self._read_cbf_image()
            self._raw_data = data

        return self._raw_data



from __future__ import annotations

import math
import os
import re
import sys
import time

from boost_adaptbx.boost.python import streambuf
from scitbx.array_family import flex

from dxtbx import IncorrectFormatError
from dxtbx.ext import is_big_endian, read_uint16, read_uint16_bs
from dxtbx.format.Format import Format


class FormatNoniusKappaCCD(Format):
    """A class for reading files produced by 1990s era Nonius Kappa CCD instruments."""

    @staticmethod
    def read_header_lines(image_path):
        """
        Adapted from Fabio KcdImage _readheader
        """

        hdr_lines = []
        end_of_headers = False
        linect = 0
        with Format.open_file(image_path, "rb") as f:
            while not end_of_headers:
                one_line = f.readline()
                linect += 1
                try:
                    one_line = one_line.decode("ASCII")
                except UnicodeDecodeError:
                    end_of_headers = True
                else:
                    if len(one_line) > 100:
                        end_of_headers = True
                if not end_of_headers:
                    if one_line.strip() == "Binned mode":
                        one_line = "Mode = Binned"
                    if "=" not in one_line and linect > 1:
                        end_of_headers = True
                    if not end_of_headers:
                        hdr_lines.append(one_line)

        return hdr_lines

    @staticmethod
    def parse_header(header_lines):
        header_dic = {}

        for l in header_lines:
            separator = l.find("=")
            if separator == -1:
                continue
            k = l[:separator].strip()
            v = l[separator + 1 :].strip()
            if k in header_dic:
                header_dic[k] = header_dic[k] + "\n" + v
            else:
                header_dic[k] = v

        return header_dic

    @staticmethod
    def understand(image_file):
        """
        Look for characteristic character sequences in the
        header.
        """
        try:
            with Format.open_file(image_file, "rb") as fh:
                tag = fh.read(1024).decode("latin-1", "replace")
        except OSError:
            return False
        if tag[0:2] != "No":
            return False
        if re.search("^Kappa-support angle", tag, re.MULTILINE) is None:
            return False
        if re.search("^Binned mode", tag, re.MULTILINE) is None:
            return False
        if re.search("^Shutter = Closed", tag, re.MULTILINE) is not None:
            return False  # ignore dark current measurement
        return True

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(str(image_file), **kwargs)

    def _start(self):
        self.headers = self.parse_header(self.read_header_lines(self._image_file))

    def _goniometer(self):
        """Return model for a kappa goniometer"""
        alpha = float(self.headers["Kappa-support angle"])
        omega = float(self.headers["Omega start"])
        kappa = float(self.headers["Kappa start"])
        phi = float(self.headers["Phi start"])
        direction = "-z"  #

        if "Phi scan range" in self.headers:
            scan_axis = "phi"
        else:
            scan_axis = "omega"

        return self._goniometer_factory.make_kappa_goniometer(
            alpha, omega, kappa, phi, direction, scan_axis
        )

    def _detector(self):
        """It appears that pixel size reported in the header does not
        account for binning, which doubles the size in both dimensions."""

        pix_fac = 0.001
        if self.headers["Mode"] == "Binned":
            pix_fac = 0.002

        pixel_size = [
            float(self.headers["pixel X-size (um)"]) * pix_fac,
            float(self.headers["pixel Y-size (um)"]) * pix_fac,
        ]

        image_size = [
            float(self.headers["X dimension"]),
            float(self.headers["Y dimension"]),
        ]

        beam_centre = (
            image_size[0] * pixel_size[0] * 0.5,
            image_size[1] * pixel_size[1] * 0.5,
        )

        gain = self.headers.get("Detector gain (ADU/count)", "1.0")
        gain = float(gain)

        return self._detector_factory.two_theta(
            sensor="CCD",
            distance=float(self.headers["Dx start"]),
            beam_centre=beam_centre,
            fast_direction="+y",  # Increasing to the right
            slow_direction="+x",  # Down, same as rotation axis
            two_theta_direction="+x",  # Same sense as omega and phi
            two_theta_angle=float(self.headers["Theta start"]) * 2.0,
            pixel_size=pixel_size,
            image_size=image_size,
            gain=gain,
            identifier=self.headers["Detector ID"],
            trusted_range=(0.0, 131070.0),  # May readout UInt16 twice
        )

    def _beam(self):
        """Return a simple model for a lab-based beam. As dxtbx
        has no laboratory target model, take the weighted values
        of the alpha1/alpha2 wavelengths."""

        a1 = float(self.headers["Alpha1"])
        a2 = float(self.headers["Alpha2"])
        wt = float(self.headers["Alpha1/Alpha2 ratio"])
        wavelength = (wt * a1 + a2) / (wt + 1.0)
        direction = [0.0, 0.0, 1.0]
        polarisation, pol_dir = self.get_beam_polarisation()
        return self._beam_factory.complex(
            sample_to_source=direction,
            wavelength=wavelength,
            polarization_plane_normal=pol_dir,
            polarization_fraction=polarisation,
        )

    def _scan(self):
        """All scan dates will be in the range 1969-2068 as per
        Python strptime. Please retire your CCD before 2069."""

        if "Phi scan range" in self.headers:
            rotax = "Phi"
        else:
            rotax = "Omega"

        osc_start = float(self.headers["%s start" % rotax])
        osc_range = float(self.headers["%s scan range" % rotax])
        exposure_time = float(self.headers["Exposure time"])

        simpletime = self.headers["Date"]
        epoch = time.mktime(time.strptime(simpletime, "%a %m/%d/%y    %I:%M:%S %p"))

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    def get_beam_polarisation(self):
        """Polarisation for single reflection off graphite 002
        monochromator as per manual. Hard-coded angles for Mo, Cu and
        Ag targets."""

        if self.headers["Target material"] == "MO":
            two_theta = 12.2  # From manual
        elif self.headers["Target material"] == "CU":
            two_theta = 26.6  # From manual
        elif self.headers["Target material"] == "AG":
            two_theta = 9.62  # Calculated

        # Assume an ideally imperfect monochromator, with the axis
        # of rotation of the monochromator in the horizontal direction
        # in a laboratory setup.

        pol_frac = 1.0 / (1 + math.cos(two_theta * math.pi / 180) ** 2)
        pol_dir = [0.0, 1.0, 0.0]
        return pol_frac, pol_dir

    def get_raw_data(self):
        """Return raw data from a Nonius CCD frame."""

        # Frame file contains a series of readouts, each pixel is
        # an unsigned little-endian 16-bit integer

        dim1 = int(self.headers["X dimension"])
        dim2 = int(self.headers["Y dimension"])
        nbReadOut = int(self.headers["Number of readouts"])
        expected_size = dim1 * dim2 * 2 * nbReadOut

        # Not clear if the image offset is present in the header,
        # therefore we seek from the end of the file.

        with self.open_file(self._image_file, "rb") as infile:
            fileSize = os.stat(self._image_file)[6]
            infile.seek(fileSize - expected_size, os.SEEK_SET)

            for i in range(nbReadOut):
                if is_big_endian():
                    raw_data = read_uint16_bs(streambuf(infile), dim1 * dim2)
                else:
                    raw_data = read_uint16(streambuf(infile), dim1 * dim2)

                if i == 0:
                    data = raw_data
                else:
                    data += raw_data

        data.reshape(flex.grid(dim2, dim1))
        return data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatNoniusKappaCCD.understand(arg))



"""
Implementation of an ImageFormat class to read TIFF format image but not -
in the first instance - actually provide a full image representation. This
is simply there to set everything up for the Mar / Rayonix CCD readers
which really will acquire the full image including header information
and generate the experimental model representations.
"""

from __future__ import annotations

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format
from dxtbx.format.FormatTIFFHelpers import (
    BIG_ENDIAN,
    LITTLE_ENDIAN,
    read_basic_tiff_header,
)


class FormatTIFF(Format):
    """An image reading class for TIFF format images i.e. those from Dectris
    and Rayonix, which start with a standard TIFF header (which is what is
    handled here) and have their own custom header following, which must
    be handled by the inheriting subclasses."""

    LITTLE_ENDIAN = LITTLE_ENDIAN
    BIG_ENDIAN = BIG_ENDIAN

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an TIFF format image, i.e. we can
        make sense of it."""

        try:
            if None in read_basic_tiff_header(image_file):
                return False
        except Exception:
            return False

        return True

    @staticmethod
    def get_tiff_header(image_file):
        """Pun to get to the image header etc."""

        width, height, depth, header, order = read_basic_tiff_header(image_file)
        header_bytes = FormatTIFF.open_file(image_file, "rb").read(header)

        return width, height, depth // 8, order, header_bytes

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)

        super().__init__(image_file, **kwargs)

    def _start(self):
        """Open the image file, read the image header, copy it into memory
        for future inspection."""

        width, height, depth, header, order = read_basic_tiff_header(self._image_file)

        self._tiff_width = width
        self._tiff_height = height
        self._tiff_depth = depth // 8
        self._tiff_header_bytes = FormatTIFF.open_file(self._image_file, "rb").read(
            header
        )
        self._tiff_byte_order = order



from __future__ import annotations

from libtbx.test_utils import approx_equal

from dxtbx.imageset import ImageSetFactory


def test_multi_axis_goniometer(dials_data):
    data_dir = dials_data("misc_regression")

    imgset = ImageSetFactory.new(str(data_dir / "dials-190_01_00001.cbf"))[0]
    gonio = imgset.get_goniometer(0)
    assert approx_equal(gonio.get_fixed_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(gonio.get_setting_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))

    imgset = ImageSetFactory.new(str(data_dir / "dials-190_02_00001.cbf"))[0]
    gonio = imgset.get_goniometer(0)
    assert approx_equal(gonio.get_fixed_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(
        gonio.get_setting_rotation(),
        (1, 0, 0, 0, -0.5, 0.866, 0.0, -0.866, -0.5),
        eps=1e-4,
    )

    imgset = ImageSetFactory.new(str(data_dir / "dials-190_03_00001.cbf"))[0]
    gonio = imgset.get_goniometer(0)
    assert approx_equal(gonio.get_fixed_rotation(), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    assert approx_equal(
        gonio.get_setting_rotation(),
        (1, 0, 0, 0, 0.5, 0.866, 0.0, -0.866, 0.5),
        eps=1e-4,
    )


def test_still(dials_data):
    data_dir = dials_data("image_examples")

    imgset = ImageSetFactory.new(str(data_dir / "DLS_I04-grid_full_cbf_0005.cbf.bz2"))[
        0
    ]
    if imgset.get_scan(0):
        scan = imgset.get_scan(0)
        assert approx_equal(
            scan.get_image_oscillation(scan.get_image_range()[0]), (30.0, 0.0), eps=1e-5
        )
    beam = imgset.get_beam(0)
    beam.get_s0()
    assert approx_equal(beam.get_s0(), (-0.0, -0.0, -1.0209290454313424))



from __future__ import annotations

import collections
import copy
import errno
import importlib.metadata
import itertools
import json
import logging
import operator
import os
import pickle
import sys
from collections.abc import Callable, Generator, Iterable
from typing import Any

import natsort
from tqdm import tqdm

import dxtbx
from dxtbx.format.Format import Format
from dxtbx.format.FormatMultiImage import FormatMultiImage
from dxtbx.format.image import ImageBool, ImageDouble
from dxtbx.format.Registry import get_format_class_for_file
from dxtbx.imageset import ImageGrid, ImageSequence, ImageSet, ImageSetFactory
from dxtbx.model import (
    BeamFactory,
    CrystalFactory,
    DetectorFactory,
    Experiment,
    ExperimentList,
    GoniometerFactory,
    History,
    ProfileModelFactory,
    ScanFactory,
)
from dxtbx.sequence_filenames import (
    locate_files_matching_template_string,
    template_image_range,
    template_regex,
    template_string_number_index,
)
from dxtbx.serialize import xds
from dxtbx.serialize.filename import resolve_path
from dxtbx.util import get_url_scheme

__all__ = [
    "BeamComparison",
    "DetectorComparison",
    "ExperimentListFactory",
    "GoniometerComparison",
]


logger = logging.getLogger(__name__)

# REMOVE and inline when Python 3.10 is minimum
if sys.version_info < (3, 10):
    scaling_model_entry_points = importlib.metadata.entry_points().get(
        "dxtbx.scaling_model_ext", []
    )
else:
    scaling_model_entry_points = importlib.metadata.entry_points(
        group="dxtbx.scaling_model_ext"
    )


class InvalidExperimentListError(RuntimeError):
    """
    Indicates an error whilst validating the experiment list.

    This means that there is some structural problem that prevents the given data
    from representing a well-formed experiment list. This doesn't indicate e.g.
    some problem with the data or model consistency.
    """


class FormatChecker:
    """A helper class to speed up identifying the correct image format by first
    trying the last format that was used."""

    def __init__(self):
        """Set the format class to none."""
        self._format_class = None

    def find_format(self, filename):
        """Search the registry for the image format class.
        Where possible use the last seen format class as a prioritisation hint.
        """
        if self._format_class:
            self._format_class = get_format_class_for_file(
                filename, format_hint=self._format_class.__name__
            )
        else:
            self._format_class = get_format_class_for_file(filename)
        if self._format_class:
            logger.debug("Using %s for %s", self._format_class.__name__, filename)
        else:
            logger.debug("No format class found for %s", filename)
        return self._format_class

    def iter_groups(self, filenames):
        group_format = None
        group_fnames = []
        for filename in filenames:
            fmt = self.find_format(filename)
            if fmt == group_format:
                group_fnames.append(filename)
            else:
                if group_fnames:
                    yield group_format, group_fnames
                group_fnames = [filename]
                group_format = fmt
            if fmt is not None:
                logger.debug("Using %s for %s", fmt.__name__, filename)
        if group_fnames:
            yield group_format, group_fnames


class BeamComparison:
    """A class to provide simple beam comparison"""

    def __init__(
        self,
        wavelength_tolerance=1e-6,
        direction_tolerance=1e-6,
        polarization_normal_tolerance=1e-6,
        polarization_fraction_tolerance=1e-6,
    ):
        self.wavelength_tolerance = wavelength_tolerance
        self.direction_tolerance = direction_tolerance
        self.polarization_normal_tolerance = polarization_normal_tolerance
        self.polarization_fraction_tolerance = polarization_fraction_tolerance

    def __call__(self, a, b):
        if a is None and b is None:
            return True
        return a.is_similar_to(
            b,
            wavelength_tolerance=self.wavelength_tolerance,
            direction_tolerance=self.direction_tolerance,
            polarization_normal_tolerance=self.polarization_normal_tolerance,
            polarization_fraction_tolerance=self.polarization_fraction_tolerance,
        )


class DetectorComparison:
    """A class to provide simple detector comparison"""

    def __init__(
        self, fast_axis_tolerance=1e-6, slow_axis_tolerance=1e-6, origin_tolerance=1e-6
    ):
        self.fast_axis_tolerance = fast_axis_tolerance
        self.slow_axis_tolerance = slow_axis_tolerance
        self.origin_tolerance = origin_tolerance

    def __call__(self, a, b):
        if a is None and b is None:
            return True
        return a.is_similar_to(
            b,
            fast_axis_tolerance=self.fast_axis_tolerance,
            slow_axis_tolerance=self.slow_axis_tolerance,
            origin_tolerance=self.origin_tolerance,
        )


class GoniometerComparison:
    """A class to provide simple goniometer comparison"""

    def __init__(
        self,
        rotation_axis_tolerance=1e-6,
        fixed_rotation_tolerance=1e-6,
        setting_rotation_tolerance=1e-6,
    ):
        self.rotation_axis_tolerance = rotation_axis_tolerance
        self.fixed_rotation_tolerance = fixed_rotation_tolerance
        self.setting_rotation_tolerance = setting_rotation_tolerance

    def __call__(self, a, b):
        if a is None and b is None:
            return True
        elif a is None or b is None:
            return False
        return a.is_similar_to(
            b,
            rotation_axis_tolerance=self.rotation_axis_tolerance,
            fixed_rotation_tolerance=self.fixed_rotation_tolerance,
            setting_rotation_tolerance=self.setting_rotation_tolerance,
        )


class ExperimentListDict:
    """A helper class for serializing the experiment list to dictionary (needed
    to save the experiment list to JSON format."""

    def __init__(self, obj, check_format=True, directory=None):
        """Initialise. Copy the dictionary."""
        # Basic check: This is a dict-like object. This can happen if e.g. we
        # were passed a DataBlock list instead of an ExperimentList dictionary
        if isinstance(obj, list) or not hasattr(obj, "get"):
            raise InvalidExperimentListError(f"Expected dictionary, not {type(obj)}")

        self._obj = copy.deepcopy(obj)
        self._check_format = check_format
        self._directory = directory

        # If this doesn't claim to be an ExperimentList, don't even try
        if self._obj.get("__id__") != "ExperimentList":
            raise InvalidExperimentListError(
                "Expected __id__ 'ExperimentList', but found {}".format(
                    repr(self._obj.get("__id__"))
                )
            )

        # Extract lists of models referenced by experiments
        # Go through all the imagesets and make sure the dictionary
        # references by an index rather than a file path.
        self._lookups = {
            model: self._extract_models(model, function)
            for model, function in (
                ("beam", BeamFactory.from_dict),
                ("detector", DetectorFactory.from_dict),
                ("goniometer", GoniometerFactory.from_dict),
                ("scan", ScanFactory.from_dict),
                ("crystal", CrystalFactory.from_dict),
                ("profile", ProfileModelFactory.from_dict),
                ("imageset", lambda x: x),
                ("scaling_model", self._scaling_model_from_dict),
            )
        }

    def _extract_models(self, name, from_dict):
        """
        Helper function. Extract the models.

        if name == imageset: Extract imageset objects from the source.

        This function does resolving of an (old) method of imageset lookup
        e.g. it was valid to have a string as the imageset value in an
        experiment instead of an int - in which case the imageset was
        loaded from the named file in the target directory.

        If any experiments point to a file in this way, the imageset is
        loaded and the experiment is rewritten with an integer pointing
        to the new ImageSet in the returned list.

        Returns:
                The ordered list of serialized-ImageSet dictionaries
                that the Experiment list points to.
        """

        # Extract all the model list
        mlist = self._obj.get(name, [])

        # Convert the model from dictionary to concreate
        # python class for the model.
        mlist = [from_dict(d) for d in mlist]

        # Dictionaries for file mappings
        mmap = {}

        # For each experiment, check the model is not specified by
        # a path, if it is then get the dictionary of the model
        # and insert it into the list. Replace the path reference
        # with an index
        for eobj in self._obj["experiment"]:
            value = eobj.get(name)
            if value is None:
                continue
            elif isinstance(value, str):
                if value not in mmap:
                    mmap[value] = len(mlist)
                    mlist.append(
                        from_dict(_experimentlist_from_file(value, self._directory))
                    )
                eobj[name] = mmap[value]
            elif not isinstance(value, int):
                raise TypeError("expected int or str, got %s" % type(value))

        return mlist

    def _load_pickle_path(
        self, imageset_data: dict, param: str
    ) -> tuple[str | None, Any]:
        """
        Read a filename from an imageset dict and load if available. This is
        used to load mask, gain, pedestal and offset maps. In some situations
        (such as tests) these files are not available, in which case the
        filename is kept but the data is None.

        Args:
            imageset_data: The dictionary holding imageset information
            param: The key name to lookup in the imageset dictionary

        Returns:
            A tuple of (filename, data) where data has been loaded from
            the pickle file, or is None if the file is inaccessible. If there
            is no key entry then ("", None) is returned.
        """
        if param not in imageset_data:
            return "", None

        filename = resolve_path(imageset_data[param], directory=self._directory)
        data = None
        if filename:
            try:
                with open(filename, "rb") as fh:
                    data = pickle.load(fh, encoding="bytes")
            except OSError:
                pass
        else:
            filename = ""

        return filename, data

    def _imageset_from_imageset_data(self, imageset_data, models):
        """Make an imageset from imageset_data - help with refactor decode."""
        assert imageset_data is not None
        if "params" in imageset_data:
            format_kwargs = imageset_data["params"]
        else:
            format_kwargs = {}

        beam = models["beam"]
        detector = models["detector"]
        goniometer = models["goniometer"]
        scan = models["scan"]

        # Load the external lookup data
        mask_filename, mask = self._load_pickle_path(imageset_data, "mask")
        gain_filename, gain = self._load_pickle_path(imageset_data, "gain")
        pedestal_filename, pedestal = self._load_pickle_path(imageset_data, "pedestal")
        dx_filename, dx = self._load_pickle_path(imageset_data, "dx")
        dy_filename, dy = self._load_pickle_path(imageset_data, "dy")

        # If dx, dy maps are expected then they must be loaded even when
        # self._check_format == False, because they affect the operation of
        # programs (dials.index, dials.refine) that do not need the image data.
        if (dx_filename or dy_filename) and not all((dx, dx)):
            raise RuntimeError(
                f"dx ({dx_filename}) and dy ({dy_filename}) maps are expected"
            )

        if imageset_data["__id__"] == "ImageSet":
            imageset = self._make_stills(imageset_data, format_kwargs=format_kwargs)
        elif imageset_data["__id__"] == "ImageGrid":
            imageset = self._make_grid(imageset_data, format_kwargs=format_kwargs)
        elif (
            imageset_data["__id__"] == "ImageSequence"
            or imageset_data["__id__"] == "ImageSweep"
        ):
            imageset = self._make_sequence(
                imageset_data,
                beam=beam,
                detector=detector,
                goniometer=goniometer,
                scan=scan,
                format_kwargs=format_kwargs,
            )
        elif imageset_data["__id__"] == "MemImageSet":
            imageset = self._make_mem_imageset(imageset_data)
        else:
            raise RuntimeError("Unknown imageset type")

        if imageset is not None:
            # Set the external lookup
            if mask is None:
                mask = ImageBool()
            else:
                mask = ImageBool(mask)
            if gain is None:
                gain = ImageDouble()
            else:
                gain = ImageDouble(gain)
            if pedestal is None:
                pedestal = ImageDouble()
            else:
                pedestal = ImageDouble(pedestal)
            if dx is None:
                dx = ImageDouble()
            else:
                dx = ImageDouble(dx)
            if dy is None:
                dy = ImageDouble()
            else:
                dy = ImageDouble(dy)

            if not imageset.external_lookup.mask.data.empty():
                if not mask.empty():
                    mask = tuple(m.data() for m in mask)
                    for m1, m2 in zip(mask, imageset.external_lookup.mask.data):
                        m1 &= m2.data()
                    imageset.external_lookup.mask.data = ImageBool(mask)
            else:
                imageset.external_lookup.mask.data = mask
            imageset.external_lookup.mask.filename = mask_filename
            imageset.external_lookup.gain.data = gain
            imageset.external_lookup.gain.filename = gain_filename
            imageset.external_lookup.pedestal.data = pedestal
            imageset.external_lookup.pedestal.filename = pedestal_filename
            imageset.external_lookup.dx.data = dx
            imageset.external_lookup.dx.filename = dx_filename
            imageset.external_lookup.dy.data = dy
            imageset.external_lookup.dy.filename = dy_filename

            # Update the imageset models
            if isinstance(imageset, ImageSequence):
                imageset.set_beam(beam)
                imageset.set_detector(detector)
                imageset.set_goniometer(goniometer)
                imageset.set_scan(scan)
            elif isinstance(imageset, (ImageSet, ImageGrid)):
                for i in range(len(imageset)):
                    imageset.set_beam(beam, i)
                    imageset.set_detector(detector, i)
                    imageset.set_goniometer(goniometer, i)
                    imageset.set_scan(scan, i)
            imageset.update_detector_px_mm_data()

        return imageset

    def decode(self):
        """Decode the dictionary into a list of experiments."""
        # Extract all the experiments - first find all scans belonging to
        # same imageset

        eobj_scan = {}

        for eobj in self._obj["experiment"]:
            if self._lookup_model("imageset", eobj) is None:
                continue
            imageset_ref = eobj.get("imageset")
            scan = self._lookup_model("scan", eobj)

            if imageset_ref in eobj_scan:
                # if there is no scan, or scan is identical, move on, else
                # make a scan which encompasses both scans
                if not scan or scan == eobj_scan[imageset_ref]:
                    continue
                i = eobj_scan[imageset_ref].get_image_range()
                j = scan.get_image_range()
                if i[1] + 1 == j[0]:
                    eobj_scan[imageset_ref] += scan
                else:
                    # make a new bigger scan
                    o = eobj_scan[imageset_ref].get_oscillation()
                    s = scan.get_oscillation()
                    assert abs(o[1] - (s[1])) < 1e-7
                    scan = copy.deepcopy(scan)
                    scan.set_image_range((min(i[0], j[0]), max(i[1], j[1])))
                    scan.set_oscillation((min(o[0], s[0]), o[1]))
                    eobj_scan[imageset_ref] = scan
            else:
                eobj_scan[imageset_ref] = copy.deepcopy(scan)

        # Map of imageset/scan pairs
        imagesets = {}

        # For every experiment, use the given input to create
        # a sensible experiment.
        el = ExperimentList()
        for eobj in self._obj["experiment"]:
            # Get the models
            identifier = eobj.get("identifier", "")
            beam = self._lookup_model("beam", eobj)
            detector = self._lookup_model("detector", eobj)
            goniometer = self._lookup_model("goniometer", eobj)
            scan = self._lookup_model("scan", eobj)
            crystal = self._lookup_model("crystal", eobj)
            profile = self._lookup_model("profile", eobj)
            scaling_model = self._lookup_model("scaling_model", eobj)

            models = {
                "beam": beam,
                "detector": detector,
                "goniometer": goniometer,
                "scan": scan,
                "crystal": crystal,
                "profile": profile,
                "scaling_model": scaling_model,
            }

            imageset_ref = eobj.get("imageset")

            # If not already cached, load this imageset
            if imageset_ref not in imagesets:
                imageset_data = self._lookup_model("imageset", eobj)
                if imageset_data is not None:
                    # Create the imageset from the input data
                    models["scan"] = eobj_scan[imageset_ref]
                    imageset = self._imageset_from_imageset_data(imageset_data, models)
                    imagesets[imageset_ref] = imageset
                else:
                    # Even if we have an empty entry, this counts as a load
                    imagesets[imageset_ref] = None

            # Append the experiment
            el.append(
                Experiment(
                    imageset=imagesets[imageset_ref],
                    beam=beam,
                    detector=detector,
                    goniometer=goniometer,
                    scan=scan,
                    crystal=crystal,
                    profile=profile,
                    scaling_model=scaling_model,
                    identifier=identifier,
                )
            )

        # Add the history to the experiments if it exists
        if "history" in self._obj:
            history = History(self._obj["history"])
            for expt in el:
                expt.history = history

        return el

    def _make_mem_imageset(self, imageset):
        """Can't make a mem imageset from dict."""
        return None

    def _make_stills(self, imageset, format_kwargs=None):
        """Make a still imageset."""
        filenames = [
            resolve_path(p, directory=self._directory) if not get_url_scheme(p) else p
            for p in imageset["images"]
        ]
        indices = None
        if "single_file_indices" in imageset:
            indices = imageset["single_file_indices"]
            assert len(indices) == len(filenames)
        return ImageSetFactory.make_imageset(
            filenames,
            None,
            check_format=self._check_format,
            single_file_indices=indices,
            format_kwargs=format_kwargs,
        )

    def _make_grid(self, imageset, format_kwargs=None):
        """Make a still imageset."""
        grid_size = imageset["grid_size"]
        return ImageGrid.from_imageset(
            self._make_stills(imageset, format_kwargs=format_kwargs), grid_size
        )

    def _make_sequence(
        self,
        imageset,
        beam=None,
        detector=None,
        goniometer=None,
        scan=None,
        format_kwargs=None,
    ):
        """Make an image sequence."""
        # Get the template format
        template = resolve_path(imageset["template"], directory=self._directory)

        # Get the number of images (if no scan is given we'll try
        # to find all the images matching the template
        if scan is None:
            i0, i1 = template_image_range(template)
        else:
            i0, i1 = scan.get_image_range()

        format_class = None
        if self._check_format is False:
            if "single_file_indices" in imageset:
                format_class = FormatMultiImage

        # Make a sequence from the input data
        return ImageSetFactory.make_sequence(
            template,
            list(range(i0, i1 + 1)),
            format_class=format_class,
            check_format=self._check_format,
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            format_kwargs=format_kwargs,
        )

    def _lookup_model(self, name, experiment_dict):
        """
        Find a model by looking up its index from a dictionary

        Args:
            name (str): The model name e.g. 'beam', 'detector'
            experiment_dict (Dict[str, int]):
                The experiment dictionary. experiment_dict[name] must
                exist and be not None to retrieve a model. If this key
                exists, then there *must* be an item with this index
                in the ExperimentListDict internal model stores.

        Returns:
            Optional[Any]:
                A model by looking up the index pointed to by
                experiment_dict[name]. If not present or empty,
                then None is returned.
        """
        if experiment_dict.get(name) is None:
            return None
        return self._lookups[name][experiment_dict[name]]

    @staticmethod
    def _scaling_model_from_dict(obj):
        """Get the scaling model from a dictionary."""
        for entry_point in scaling_model_entry_points:
            if entry_point.name == obj["__id__"]:
                return entry_point.load().from_dict(obj)


def _experimentlist_from_file(filename, directory=None):
    """Load a model dictionary from a file."""
    filename = resolve_path(filename, directory=directory)
    try:
        with open(filename) as infile:
            return json.load(infile)
    except OSError:
        raise OSError("unable to read file, %s" % filename)


class ExperimentListFactory:
    """A class to help instantiate experiment lists."""

    @staticmethod
    def from_args(
        args: list[str], unhandled: list[str] | None = None, check_format: bool = True
    ) -> ExperimentList:
        """Try to load serialised experiments from any recognised format."""

        # Create a list for unhandled arguments
        if unhandled is None:
            unhandled = []

        experiments = ExperimentList()

        # Try to load from serialized formats
        for filename in args:
            try:
                experiments.extend(
                    ExperimentListFactory.from_serialized_format(
                        filename, check_format=check_format
                    )
                )
                logger.debug(f"Loaded experiments from {filename}")
            except Exception as e:
                logger.debug(f"Could not load experiments from {filename}: {e}")
                unhandled.append(filename)
                raise

        return experiments

    @staticmethod
    def from_filenames(
        filenames,
        unhandled=None,
        compare_beam=None,
        compare_detector=None,
        compare_goniometer=None,
        scan_tolerance=None,
        format_kwargs=None,
        load_models=True,
    ) -> ExperimentList:
        """Create a list of data blocks from a list of directory or file names."""
        experiments = ExperimentList()

        # Cast filenames to a list from whatever iterator they are
        filenames = list(filenames)

        # Process each file given by this path list
        to_process = _openingpathiterator(filenames)
        find_format = FormatChecker()

        format_groups = collections.defaultdict(list)
        if format_kwargs is None:
            format_kwargs = {}

        if os.isatty and len(filenames) > 1 and "DIALS_NOBANNER" not in os.environ:
            filename_iter = tqdm(to_process, total=len(filenames), file=sys.stdout)
        else:
            filename_iter = to_process

        for filename in filename_iter:
            # We now have a file, pre-opened by Format.open_file (therefore
            # cached). Determine its type, and prepare to put into a group
            format_class = find_format.find_format(filename)

            # Verify this makes sense
            if not format_class:
                # No format class found?
                logger.debug("Could not determine format for %s", filename)
                if unhandled is not None:
                    unhandled.append(filename)
            elif format_class.is_abstract():
                logger.debug(
                    f"Image file {filename} appears to be a '{format_class.__name__}', but this is an abstract Format"
                )
                # Invalid format class found?
                if unhandled is not None:
                    unhandled.append(filename)
            elif issubclass(format_class, FormatMultiImage):
                imageset = format_class.get_imageset(
                    os.path.abspath(filename), format_kwargs=format_kwargs
                )
                format_groups[format_class].append(imageset)
                logger.debug("Loaded file: %s", filename)
            else:
                format_object = format_class(filename, **format_kwargs)
                meta = ImageMetadataRecord.from_format(format_object)
                assert meta.filename == filename

                # Add this entry to our table of formats
                format_groups[format_class].append(meta)
                logger.debug("Loaded metadata of file: %s", filename)

        # Now, build experiments from these files. Duplicating the logic of
        # the previous implementation:
        # - FormatMultiImage files each have their own ImageSet
        # - Every set of images forming a scan goes into its own ImageSequence
        # - Any consecutive still frames that share any metadata with the
        #   previous still fram get collected into one ImageSet

        all_tof = False
        for format_class, records in format_groups.items():
            for i in records:
                try:  # records can be ImageMetadataRecord or ImageSequence
                    scan = i.get_scan()
                    if scan is not None and scan.has_property("time_of_flight"):
                        all_tof = True
                    elif all_tof:
                        raise RuntimeError(
                            "Cannot process mix of ToF and non ToF experiments"
                        )
                except AttributeError:
                    if all_tof:
                        raise RuntimeError(
                            "Cannot process mix of ToF and non ToF experiments"
                        )

        # Treat each format as a separate block of data
        for format_class, records in format_groups.items():
            if issubclass(format_class, FormatMultiImage):
                if all_tof:
                    _merge_sequence_model_metadata(
                        records,
                        compare_beam=compare_beam,
                        compare_detector=compare_detector,
                        compare_goniometer=compare_goniometer,
                    )

                for imageset in records:
                    experiments.extend(
                        ExperimentListFactory.from_imageset_and_crystal(
                            imageset, crystal=None, load_models=load_models
                        )
                    )
                continue

            # Merge any consecutive and identical metadata together
            _merge_model_metadata(
                records,
                compare_beam=compare_beam,
                compare_detector=compare_detector,
                compare_goniometer=compare_goniometer,
            )
            records = _merge_scans(records, scan_tolerance=scan_tolerance)
            imagesets = list(
                _convert_to_imagesets(records, format_class, format_kwargs)
            )

            assert imagesets, "Got no imagesets when constructing ExperimentList?"
            for imageset in imagesets:
                experiments.extend(
                    ExperimentListFactory.from_imageset_and_crystal(
                        imageset, crystal=None, load_models=load_models
                    )
                )

        return experiments

    @staticmethod
    def from_imageset_and_crystal(imageset: ImageSet | None, crystal, load_models=True):
        """Load an experiment list from an imageset and crystal."""
        if isinstance(imageset, ImageSequence):
            return ExperimentListFactory.from_sequence_and_crystal(
                imageset, crystal, load_models
            )
        else:
            return ExperimentListFactory.from_stills_and_crystal(
                imageset, crystal, load_models
            )

    @staticmethod
    def from_sequence_and_crystal(imageset, crystal, load_models=True):
        """Create an experiment list from sequence and crystal."""

        assert isinstance(imageset, ImageSequence)

        experiments = ExperimentList()

        if load_models:
            # if imagesequence is still images, make one experiment for each
            # all referencing into the same image set
            if imageset.get_scan().is_still():
                for j in range(len(imageset)):
                    subset = imageset[j : j + 1]
                    experiments.append(
                        Experiment(
                            imageset=imageset,
                            beam=imageset.get_beam(),
                            detector=imageset.get_detector(),
                            goniometer=imageset.get_goniometer(),
                            scan=subset.get_scan(),
                            crystal=crystal,
                        )
                    )
            else:
                experiments.append(
                    Experiment(
                        imageset=imageset,
                        beam=imageset.get_beam(),
                        detector=imageset.get_detector(),
                        goniometer=imageset.get_goniometer(),
                        scan=imageset.get_scan(),
                        crystal=crystal,
                    )
                )

            return experiments

        else:
            return ExperimentList([Experiment(imageset=imageset, crystal=crystal)])

    @staticmethod
    def from_stills_and_crystal(imageset: ImageSet | None, crystal, load_models=True):
        """Create an experiment list from stills and crystal."""
        experiments = ExperimentList()
        if imageset is None:
            return experiments
        if load_models:
            for i in range(len(imageset)):
                experiments.append(
                    Experiment(
                        imageset=imageset[i : i + 1],
                        beam=imageset.get_beam(i),
                        detector=imageset.get_detector(i),
                        goniometer=imageset.get_goniometer(i),
                        scan=imageset.get_scan(i),
                        crystal=crystal,
                    )
                )
        else:
            for i in range(len(imageset)):
                experiments.append(
                    Experiment(imageset=imageset[i : i + 1], crystal=crystal)
                )
        return experiments

    @staticmethod
    def from_dict(obj, check_format=True, directory=None):
        """
        Load an experiment list from a dictionary.

        Args:
            obj (dict):
                Dictionary containing either ExperimentList or DataBlock
                structure.
            check_format (bool):
                If True, the file will be read to verify metadata.
            directory (str):

        Returns:
            ExperimentList: The dictionary converted
        """

        experiments = ExperimentListDict(
            obj, check_format=check_format, directory=directory
        ).decode()

        # Check the list is consistent
        assert experiments.is_consistent()

        return experiments

    @staticmethod
    def from_json(text, check_format=True, directory=None):
        """Load an experiment list from JSON."""
        return ExperimentListFactory.from_dict(
            json.loads(text),
            check_format=check_format,
            directory=directory,
        )

    @staticmethod
    def from_json_file(filename, check_format=True):
        """Load an experiment list from a json file."""
        filename = os.path.abspath(filename)
        directory = os.path.dirname(filename)
        try:
            with open(filename) as infile:
                return ExperimentListFactory.from_json(
                    infile.read(), check_format=check_format, directory=directory
                )
        except UnicodeDecodeError:
            raise InvalidExperimentListError(
                f"Cannot interpret {filename} as an ExperimentList"
            )

    @staticmethod
    def from_pickle_file(filename):
        """Decode an experiment list from a pickle file."""
        with open(filename, "rb") as infile:
            obj = pickle.load(infile)
        assert isinstance(obj, ExperimentList)
        return obj

    @staticmethod
    def from_xds(xds_inp, xds_other):
        """Generate an experiment list from XDS files."""
        # Get the sequence from the XDS files
        sequence = xds.to_imageset(xds_inp, xds_other)

        # Get the crystal from the XDS files
        crystal = xds.to_crystal(xds_other)

        # Create the experiment list
        experiments = ExperimentListFactory.from_imageset_and_crystal(sequence, crystal)

        # Set the crystal in the experiment list
        assert len(experiments) == 1

        return experiments

    @staticmethod
    def from_serialized_format(filename, check_format=True):
        """Try to load the experiment list from a serialized format."""

        if hasattr(filename, "__fspath__"):
            filename = filename.__fspath__()  # unwrap PEP-519-style objects

        return ExperimentListFactory.from_json_file(filename, check_format)

    @staticmethod
    def from_templates(templates, **kwargs):
        """Import an experiment list from templates"""
        assert "verbose" not in kwargs, "The verbose parameter has been removed"
        assert len(templates) > 0

        experiments = ExperimentList()
        find_format = FormatChecker()

        # For each template do an import
        for template in templates:
            template = os.path.normpath(template)
            filenames = sorted(locate_files_matching_template_string(template))
            if len(filenames):
                logger.debug(
                    "The following files matched the template string:\n%s",
                    "\n".join(f" {p}" for p in filenames),
                )

            # Check if we've matched any filenames
            if len(filenames) == 0:
                raise ValueError(f"Template '{template}' does not match any files")

            # Get the format from the first image
            format_class = find_format.find_format(filenames[0])

            # Verify this makes sense
            if format_class is None:
                raise ValueError(f"Image file {filenames[0]} format is unknown")
            elif format_class.is_abstract():
                raise ValueError(
                    f"Image file {filenames[0]} appears to be a '{type(format_class).__name__}', but this is an abstract Format"
                )
            else:
                image_range = kwargs.get("image_range")
                if image_range:
                    first, last = image_range
                else:
                    first, last = template_image_range(template)

                if not kwargs.get("allow_incomplete_sequences", False):
                    if "#" in template:
                        # Check all images in range are present - if allowed
                        i0, i1 = template_string_number_index(template)
                        prefix = template[:i0]
                        suffix = template[i1:]
                        all_numbers = {
                            int(f.replace(prefix, "").replace(suffix, ""))
                            for f in filenames
                        }
                        missing = set(range(first, last + 1)) - all_numbers
                        if missing:
                            raise ValueError(
                                "Missing image{} {} from imageset ({}-{})".format(
                                    "s" if len(missing) > 1 else "",
                                    ", ".join(str(x) for x in sorted(missing)),
                                    first,
                                    last,
                                )
                            )
                    else:
                        print(
                            "Warning: Using only one template file: %s. \n "
                            "`allow_incomplete_sequence` has no effect" % template
                        )

                # Read the image
                fmt = format_class(filenames[0], **(kwargs.get("format_kwargs", {})))

                # Update the image range
                image_range = (first, last)
                scan = fmt.get_scan()
                scan.set_image_range(image_range)

                # Create the sequence and experiment
                imageset = dxtbx.imageset.ImageSetFactory.make_sequence(
                    template,
                    list(range(first, last + 1)),
                    format_class,
                    fmt.get_beam(),
                    fmt.get_detector(),
                    fmt.get_goniometer(),
                    scan,
                    format_kwargs=kwargs.get("format_kwargs"),
                )
                experiments.extend(
                    ExperimentListFactory.from_imageset_and_crystal(
                        imageset,
                        crystal=None,
                        load_models=True,
                    )
                )
        return experiments


class ImageMetadataRecord:
    """Object to store metadata information.

    This is used whilst building the experiment lists. The metadata for each
    image can be read once, and then any grouping/deduplication can happen
    later, without re-opening the original file.
    """

    def __init__(
        self,
        beam: dxtbx.model.Beam | None = None,
        detector: dxtbx.model.Detector | None = None,
        goniometer: dxtbx.model.Goniometer | None = None,
        scan: dxtbx.model.Scan | None = None,
        template: str | None = None,
        filename: str | None = None,
        index: int | None = None,
    ):
        """
        Args:
            beam:       Stores a beam model
            detector:   Stores a detector model
            goniometer: Stores a goniometer model
            scan:       Stores a scan model
            filename:   The filename this record was parsed from
            template:
                The template string parsed from the filename. Usually,
                the template is only present if a scan was found and
                oscillation width was nonzero.
            index:
                The index of this file in the template. Applying the
                index to the template field should recover the filename
        """
        self.beam = beam
        self.detector = detector
        self.goniometer = goniometer
        self.scan = scan
        self.template = template
        self.filename = filename
        self.index = index

    def merge_metadata_from(
        self,
        other_record: ImageMetadataRecord,
        compare_beam: Callable = operator.__eq__,
        compare_detector: Callable = operator.__eq__,
        compare_goniometer: Callable = operator.__eq__,
    ) -> bool:
        """
        Compare two record objects and merge equivalent data.

        This method will compare (with optional functions) instance data
        for beam, detector and goniometer. If any of the metadata for
        this record is equivalent to (but a different instance from) the
        other record, then this instance will be altered to match the
        other. The function used to compare beams, detectors and
        goniometers can be customised - but by default the normal
        equality operator is used.

        Args:
            other_record:       Another metadata instance
            compare_beam:       A function to compare beams
            compare_detector:   A function to compare detectors
            compare_goniometer: A function to compare goniometers

        Returns: True if any action was taken
        """
        # Allow 'defaults' of None to work - behavior from legacy implementation
        compare_beam = compare_beam or operator.__eq__
        compare_detector = compare_detector or operator.__eq__
        compare_goniometer = compare_goniometer or operator.__eq__

        record_altered = False
        if self.beam is not other_record.beam and compare_beam(
            self.beam, other_record.beam
        ):
            self.beam = other_record.beam
            record_altered = True
        if self.detector is not other_record.detector and compare_detector(
            self.detector, other_record.detector
        ):
            self.detector = other_record.detector
            record_altered = True
        if self.goniometer is not other_record.goniometer and compare_goniometer(
            self.goniometer, other_record.goniometer
        ):
            self.goniometer = other_record.goniometer
            record_altered = True

        return record_altered

    @classmethod
    def from_format(cls, fmt: Format) -> Any:
        """
        Read metadata information from a Format instance.

        This will only pull information out of a single format instance
        while it is open - combining metadata records must be done
        separately.

        Args:
            format: The instance of the format class to read data from

        Returns:
            A new ImageMetadataRecord with the pre-read information
        """
        record = cls()
        record.filename = fmt.get_image_file()
        # Get the metadata from the format
        try:
            record.beam = fmt.get_beam()
        except Exception:
            pass
        try:
            record.detector = fmt.get_detector()
        except Exception:
            pass
        try:
            record.goniometer = fmt.get_goniometer()
        except Exception:
            pass
        try:
            record.scan = fmt.get_scan()
        except Exception:
            pass

        # Get the template and index if possible - and only if we've got a
        # recorded oscillation value
        if record.scan is not None:
            record.template, record.index = template_regex(record.filename)

        return record

    def __repr__(self):
        items = [
            ("filename", self.filename),
            ("beam", self.beam),
            ("detector", self.detector),
            ("goiometer", self.goniometer),
            ("scan", self.scan),
            ("template", self.template),
            ("index", self.index),
        ]
        itemstr = ", ".join(x + "=" + repr(y) for x, y in items)
        return "<{}{}{}>".format(type(self).__name__, " " if itemstr else "", itemstr)

    def __hash__(self):
        return hash(
            (
                self.beam,
                self.detector,
                self.goniometer,
                self.scan,
                self.template,
                self.filename,
                self.index,
            )
        )

    def __eq__(self, other):
        if not isinstance(other, ImageMetadataRecord):
            return False
        return all(
            getattr(self, attribute) == getattr(other, attribute)
            for attribute in (
                "beam",
                "detector",
                "goniometer",
                "scan",
                "template",
                "filename",
                "index",
            )
        )

    def __ne__(self, other):
        return not self == other


def _iterate_with_previous(iterable):
    """Convenience iterator to give pairs of (previous, next) items"""
    previous = None
    for val in iterable:
        yield (previous, val)
        previous = val


def _groupby_template_is_none(
    records: Iterable[ImageMetadataRecord],
) -> Generator[list[ImageMetadataRecord], None, None]:
    """Specialization of groupby that groups records by format=None"""
    for _, group in itertools.groupby(
        enumerate(records), key=lambda x: -1 if x[1].template is None else x[0]
    ):
        yield [x[1] for x in group]


def _openingpathiterator(pathnames: Iterable[str]):
    """Utility function to efficiently open all paths.

    A path is a potential file or directory.
    Each path will be opened with :meth:`dxtbx.format.Format.open_file`,
    but in order to do so each file will only be opened once, and extraneous
    use of :func:`os.stat` will be avoided.
    Any path entries that are a directory will be recursed into, once -
    any further directories found will be ignored. Any path that is not
    a file or directory, or on which IO fails for any reason, will still
    be returned.

    Args:
        pathnames: Paths to attempt to open
    """

    # Store a tuple of (recurse, pathname) to track what was root level
    paths = collections.deque((True, x) for x in natsort.natsorted(pathnames))

    while paths:
        # Get the next path from the queue
        (do_recurse, pathname) = paths.popleft()
        pathname = os.fspath(pathname)
        try:
            # Attempt to open this 'path'
            Format.open_file(pathname)
        except OSError as e:
            if e.errno == errno.EISDIR:
                if do_recurse:
                    # We've tried to open a directory. Get all the entries...
                    subdir_paths = sorted(
                        os.path.join(pathname, x) for x in os.listdir(pathname)
                    )
                    # ... and add them to our queue. Make sure not to mark for recursion
                    paths.extendleft((False, x) for x in reversed(subdir_paths))
                    logger.debug("Adding %d files from %s", len(subdir_paths), pathname)
                else:
                    logger.debug("Not adding sub-level directory entry %s", pathname)
                # Don't return directory instances
                continue
            else:
                # A non-directory-related IO error
                logger.debug("Could not import %s: %s", pathname, os.strerror(e.errno))

        yield pathname


def _merge_model_metadata(
    records: Iterable[ImageMetadataRecord],
    compare_beam: Callable | None = None,
    compare_detector: Callable | None = None,
    compare_goniometer: Callable | None = None,
):
    """
    Merge metadata between consecutive record objects.

    This will compare each record with the previous one, and make sure
    the metadata instances are shared where appropriate.

    Args:
        records:    Records for the images to merge into imagesets
        compare_beam:       The function to to compare beams
        compare_detector:   The function to compare detectors
        compare_goniometer: The function to compare goniometers
    """
    for prev, record in _iterate_with_previous(records):
        if prev is None:
            continue
        record.merge_metadata_from(
            prev,
            compare_beam=compare_beam,
            compare_detector=compare_detector,
            compare_goniometer=compare_goniometer,
        )


def _merge_sequence_model_metadata(
    records: Iterable[ImageSequence],
    compare_beam: Callable | None = None,
    compare_detector: Callable | None = None,
    compare_goniometer: Callable | None = None,
):
    record_altered = False
    for prev, record in _iterate_with_previous(records):
        if prev is None:
            continue

        record_altered = False
        record_beam = record.get_beam()
        record_detector = record.get_detector()
        record_goniometer = record.get_goniometer()
        prev_beam = prev.get_beam()
        prev_detector = prev.get_detector()
        prev_goniometer = prev.get_goniometer()

        if record_beam is not prev_beam and compare_beam(record_beam, prev_beam):
            record.set_beam(prev_beam)
            record_altered = True
        if record_detector is not prev_detector and compare_detector(
            record_detector, prev_detector
        ):
            record.set_detector(prev_detector)
            record_altered = True
        if record_goniometer is not prev_goniometer and compare_goniometer(
            record_goniometer, prev_goniometer
        ):
            record.set_goniometer(prev_goniometer)
            record_altered = True

    return record_altered


def _merge_scans(
    records: Iterable[ImageMetadataRecord], scan_tolerance: float | None = None
) -> list[ImageMetadataRecord]:
    """
    Merge consecutive scan records with identical metadata.

    The records should have previously had their model metadata merged,
    as identity will be used to compare metadata identity at this stage.

    Args:
        records:        Records to merge
        scan_tolerance: Fraction of oscillation range to tolerate
                        when merging scan records

    Returns:
        A (potentially shorter) list of records with scans merged
    """
    merged_records = []
    logger.debug("Merging scans")
    for prev, record in _iterate_with_previous(records):
        # The first record always gets recorded
        if prev is None:
            merged_records.append(record)
            logger.debug("  Saving initial record %s", record)
            continue
        # Compare metadata instances
        same_metadata = [
            prev.beam is record.beam,
            prev.detector is record.detector,
            prev.goniometer is record.goniometer,
        ]
        # Condition for combining:
        # - All metadata must match
        # - Previous record must be templated
        # - This record must be templated
        if (
            all(same_metadata)
            and prev.template is not None
            and record.template is not None
        ):
            # Attempt to append to scan
            try:
                if scan_tolerance is None:
                    prev.scan.append(record.scan)
                else:
                    prev.scan.append(record.scan, scan_tolerance=scan_tolerance)
            except RuntimeError as e:
                logger.debug(
                    "  Failed to merge record %s with previous - writing new scan",
                    str(e),
                )
            else:
                # If we appended, then we don't need to keep this record's scan
                record.scan = prev.scan
                logger.debug("  Appended record %s to previous", record)
                continue
        merged_records.append(record)

    logger.debug("Result of merging record scans: %d records", len(merged_records))
    return merged_records


def _convert_to_imagesets(
    records: Iterable[ImageMetadataRecord],
    format_class: type[Format],
    format_kwargs: dict | None = None,
) -> Generator[dxtbx.imageset.ImageSet, None, None]:
    """
    Convert records into imagesets.

    The records should have been metadata- and scan-merged by this point.
    Rules:
    - Any groups of template=None where any of the metadata objects
      are shared, go into a single imageset
    - Anything with a template goes into a single sequence

    Args:
        records: The records to convert
        format_class: The format class for the data in this record
        format_kwargs: Any format configuration arguments to pass
            to the format imageset creator.

    Returns:
        Imagesets representing the records
    """

    # Iterate over images/sets such that template=None are clustered
    for setgroup in _groupby_template_is_none(records):
        if setgroup[0].template is not None:
            # If we have a template, then it's a sequence
            assert len(setgroup) == 1, "Got group of metadata records in template?"
            logger.debug("Creating Imagesequence from %s", setgroup[0].template)
            yield _create_imagesequence(setgroup[0], format_class, format_kwargs)
        else:
            # Without a template, it was never identified as a sequence, so an imageset
            logger.debug("Creating ImageSet from %d files", len(setgroup))
            yield _create_imageset(setgroup, format_class, format_kwargs)


def _create_imageset(
    records: Iterable[ImageMetadataRecord],
    format_class: type[Format],
    format_kwargs: dict | None = None,
) -> dxtbx.imageset.ImageSet:
    """
    Create an ImageSet object from a set of single-image records.

    Args:
        records: Single-image metadata records to merge into a single imageset
        format_class: The format class object for these image records
        format_kwargs: Extra arguments to pass to the format class when
            creating an ImageSet

    Returns:
        An imageset for all the image records
    """
    records = list(records)
    # Nothing here should have been assigned a template parameter
    assert all(x.template is None for x in records)
    # Everything should have a filename
    assert all(x.filename for x in records)
    # Extract the filenames from the records
    filenames = [
        x.filename if get_url_scheme(x.filename) else os.path.abspath(x.filename)
        for x in records
        if x.filename
    ]
    # Create the imageset
    imageset = dxtbx.imageset.ImageSetFactory.make_imageset(
        filenames, format_class, format_kwargs=format_kwargs, check_format=False
    )
    # Update all of the metadata for each record
    for i, r in enumerate(records):
        imageset.set_beam(r.beam, i)
        imageset.set_detector(r.detector, i)
        imageset.set_goniometer(r.goniometer, i)
        imageset.set_scan(r.scan, i)
    return imageset


def _create_imagesequence(
    record: ImageMetadataRecord,
    format_class: type[Format],
    format_kwargs: dict | None = None,
) -> dxtbx.imageset.ImageSequence:
    """
    Create an ImageSequence object from a single rotation data image.

    Args:
        record: Single-image metadata records to merge into a single imageset
        format_class: The format class object for these image records
        format_kwargs: Extra arguments to pass to the format class when
            creating an ImageSet

    Returns:
        An imageset representing the sequence of data
    """
    assert record.scan
    assert record.template
    index_start, index_end = record.scan.get_image_range()
    # Create the sequence
    sequence = dxtbx.imageset.ImageSetFactory.make_sequence(
        template=os.path.abspath(record.template),
        indices=list(range(index_start, index_end + 1)),
        format_class=format_class,
        beam=record.beam,
        detector=record.detector,
        goniometer=record.goniometer,
        scan=record.scan,
        format_kwargs=format_kwargs,
        # check_format=False,
    )
    return sequence



from __future__ import annotations

import math
import sys

import dials.util
from dials.util.options import OptionParser

from libtbx.phil import parse
from libtbx.test_utils import approx_equal
from scitbx.array_family import flex
from scitbx.math.superpose import least_squares_fit
from scitbx.matrix import col
from serialtbx.detector import iterate_detector_at_level

import dxtbx.util
from dxtbx.model.experiment_list import ExperimentListFactory

help_message = """
This program is used to superpose a moving detector onto a reference detector
  dxtbx.detector_superpose reference.json moving.json
"""

# Create the phil parameters
phil_scope = parse(
    """
reference_experiments = None
  .type = str
  .help = Experiments list with reference detector
moving_experiments = None
  .type = str
  .help = Experiments list with moving detector
output_experiments = superposed_experiments.json
  .type = str
  .help = Output file name with superposed detector
panel_list = None
  .type = ints
  .help = List of panels to use as reference. Use all if set to None.
apply_at_hierarchy_level = None
  .type = int
  .help = If None, apply shift at panel level.  If not None, apply shift \
          at specified hierarchy level.
fit_target = *corners centers
  .type = choice
  .help = Corners: perform superpose using corners of panels. Centers:\
          perform superpose using centers of panels (requires at least 3\
          panels
repeat_until_converged = True
  .type = bool
  .help = If True, do rounds of fitting until the angle of rotation and \
          magnitude of translation stop changing.
""",
    process_includes=True,
)


@dials.util.show_mail_handle_errors()
def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    usage = "usage: dxtbx.detector_superpose reference.json moving.json "
    parser = OptionParser(
        usage=usage,
        sort_options=True,
        phil=phil_scope,
        check_format=False,
        epilog=help_message,
    )
    params, options = parser.parse_args(args, show_diff_phil=True)

    reference_experiments = ExperimentListFactory.from_json_file(
        params.reference_experiments, check_format=False
    )
    if len(reference_experiments.detectors()) != 1:
        sys.exit("Error: Please ensure reference has only 1 detector model")
    reference = reference_experiments.detectors()[0]

    moving_experiments = ExperimentListFactory.from_json_file(
        params.moving_experiments, check_format=False
    )
    if len(moving_experiments.detectors()) != 1:
        sys.exit("Error: Please ensure moving has only 1 detector model")
    moving = moving_experiments.detectors()[0]

    # Get list of panels to compare
    if params.panel_list is None or len(params.panel_list) == 0:
        assert len(reference) == len(moving), "Detectors not same length"
        panel_ids = list(range(len(reference)))
    else:
        max_p_id = max(params.panel_list)
        assert max_p_id < len(reference), (
            "Reference detector must be at least %d panels long given the panel list"
            % (max_p_id + 1)
        )
        assert max_p_id < len(moving), (
            "Moving detector must be at least %d panels long given the panel list"
            % (max_p_id + 1)
        )
        panel_ids = params.panel_list

    if params.fit_target == "centers":
        assert len(panel_ids) >= 3, (
            "When using centers as target for superpose, detector needs at least 3 panels"
        )

    def rmsd_from_centers(a, b):
        assert len(a) == len(b)
        assert len(a) % 4 == len(b) % 4 == 0
        ca = flex.vec3_double()
        cb = flex.vec3_double()
        for i in range(len(a) // 4):
            ca.append(a[i : i + 4].mean())
            cb.append(b[i : i + 4].mean())
        return 1000 * math.sqrt((ca - cb).sum_sq() / len(ca))

    cycles = 0
    while True:
        cycles += 1

        # Treat panels as a list of 4 sites (corners) or 1 site (centers) for use with lsq superpose
        reference_sites = flex.vec3_double()
        moving_sites = flex.vec3_double()
        for panel_id in panel_ids:
            for detector, sites in zip(
                [reference, moving], [reference_sites, moving_sites]
            ):
                panel = detector[panel_id]
                size = panel.get_image_size()
                corners = flex.vec3_double(
                    [
                        panel.get_pixel_lab_coord(point)
                        for point in [
                            (0, 0),
                            (0, size[1] - 1),
                            (size[0] - 1, size[1] - 1),
                            (size[0] - 1, 0),
                        ]
                    ]
                )
                if params.fit_target == "corners":
                    sites.extend(corners)
                elif params.fit_target == "centers":
                    sites.append(corners.mean())

        # Compute super position
        rmsd = 1000 * math.sqrt(
            (reference_sites - moving_sites).sum_sq() / len(reference_sites)
        )
        print("RMSD before fit: %.1f microns" % rmsd)
        if params.fit_target == "corners":
            rmsd = rmsd_from_centers(reference_sites, moving_sites)
            print("RMSD of centers before fit: %.1f microns" % rmsd)
        lsq = least_squares_fit(reference_sites, moving_sites)
        rmsd = 1000 * math.sqrt(
            (reference_sites - lsq.other_sites_best_fit()).sum_sq()
            / len(reference_sites)
        )
        print("RMSD of fit: %.1f microns" % rmsd)
        if params.fit_target == "corners":
            rmsd = rmsd_from_centers(reference_sites, lsq.other_sites_best_fit())
            print("RMSD of fit of centers: %.1f microns" % rmsd)
        (
            angle,
            axis,
        ) = lsq.r.r3_rotation_matrix_as_unit_quaternion().unit_quaternion_as_axis_and_angle(
            deg=True
        )
        print(
            "Axis and angle of rotation: (%.3f, %.3f, %.3f), %.2f degrees"
            % (axis[0], axis[1], axis[2], angle)
        )
        print(
            "Translation (x, y, z, in microns): (%.3f, %.3f, %.3f)"
            % (1000 * lsq.t).elems
        )

        # Apply the shifts
        if params.apply_at_hierarchy_level is None:
            iterable = moving
        else:
            iterable = iterate_detector_at_level(
                moving.hierarchy(), level=params.apply_at_hierarchy_level
            )

        for group in iterable:
            fast = col(group.get_fast_axis())
            slow = col(group.get_slow_axis())
            ori = col(group.get_origin())

            group.set_frame(lsq.r * fast, lsq.r * slow, (lsq.r * ori) + lsq.t)

        if not params.repeat_until_converged:
            break

        if approx_equal(angle, 0.0, out=None) and approx_equal(
            (1000 * lsq.t).length(), 0.0, out=None
        ):
            print("Converged after", cycles, "cycles")
            break
        else:
            print("Movement not close to zero, repeating fit")
            print()

    moving_experiments.as_file(params.output_experiments)

    moved_sites = flex.vec3_double()
    for panel_id in panel_ids:
        panel = moving[panel_id]
        size = panel.get_image_size()
        corners = flex.vec3_double(
            [
                panel.get_pixel_lab_coord(point)
                for point in [
                    (0, 0),
                    (0, size[1] - 1),
                    (size[0] - 1, size[1] - 1),
                    (size[0] - 1, 0),
                ]
            ]
        )
        if params.fit_target == "corners":
            moved_sites.extend(corners)
        elif params.fit_target == "centers":
            moved_sites.append(corners.mean())

    # Re-compute RMSD after moving detector components
    rmsd = 1000 * math.sqrt(
        (reference_sites - moved_sites).sum_sq() / len(reference_sites)
    )
    print("RMSD of fit after movement: %.1f microns" % rmsd)
    if params.fit_target == "corners":
        rmsd = rmsd_from_centers(reference_sites, moved_sites)
        print("RMSD of fit of centers after movement: %.1f microns" % rmsd)

    if params.panel_list is not None:
        reference_sites = flex.vec3_double()
        moved_sites = flex.vec3_double()
        for panel_id in range(len(reference)):
            for detector, sites in zip(
                [reference, moving], [reference_sites, moved_sites]
            ):
                panel = detector[panel_id]
                size = panel.get_image_size()
                corners = flex.vec3_double(
                    [
                        panel.get_pixel_lab_coord(point)
                        for point in [
                            (0, 0),
                            (0, size[1] - 1),
                            (size[0] - 1, size[1] - 1),
                            (size[0] - 1, 0),
                        ]
                    ]
                )
                if params.fit_target == "corners":
                    sites.extend(corners)
                elif params.fit_target == "centers":
                    sites.append(corners.mean())
        # Re-compute RMSD for full detector after moving detector components
        rmsd = 1000 * math.sqrt(
            (reference_sites - moved_sites).sum_sq() / len(reference_sites)
        )
        print("RMSD of whole detector fit after movement: %.1f microns" % rmsd)
        if params.fit_target == "corners":
            rmsd = rmsd_from_centers(reference_sites, moved_sites)
            print(
                "RMSD of whole detector fit of centers after movement: %.1f microns"
                % rmsd
            )


if __name__ == "__main__":
    run()



"""
Image reading tests against image_examples data
"""

from __future__ import annotations

import pytest

import scitbx.matrix
from rstbx.slip_viewer.slip_viewer_image_factory import SlipViewerImageFactory
from scitbx.array_family import flex

import dxtbx.format.Registry
from dxtbx.util import get_url_scheme

try:
    import h5py
except ModuleNotFoundError:
    h5py = None

try:
    import cbflib_adaptbx

    # Check this isn't a namespace package
    if hasattr(cbflib_adaptbx, "__path__") and cbflib_adaptbx.__file__ is None:
        cbflib_adaptbx = None
except ModuleNotFoundError:
    # Detectorbase for CBF won't work
    cbflib_adaptbx = None

from . import imagelist

_files = imagelist.image_examples


@pytest.mark.parametrize("test_image", _files)
def test_detectorbase(test_image):
    if not h5py and test_image.endswith((".h5", ".nxs")):
        pytest.skip("could not import 'h5py'")

    if test_image.endswith(".cbf") and not cbflib_adaptbx:
        pytest.skip("No cbflib_adaptbx: CBF Detectorbase is not available")

    if test_image.endswith("bz2") or test_image.endswith("gz"):
        # Skip compressed files
        pytest.skip("Skip compressed files as detectorbase cannot be instantiated")

    format_instance = dxtbx.format.Registry.get_format_class_for_file(test_image)
    print("Reading", test_image)
    print("Format:", format_instance)
    assert format_instance, "no matching format class found"
    instance = format_instance(test_image)

    try:
        detectorbase = instance.get_detectorbase()
    except NotImplementedError:
        pytest.skip("Skipping image with no detectorbase")

    if not detectorbase:
        pytest.skip("Skipping image with no detectorbase")

    imgfactory = SlipViewerImageFactory(test_image)
    try:
        imgfactory.read()
    except AttributeError:
        pytest.skip("Skip case where SlipViewerImageFactory cannot read")

    print("  Detectorbase:", instance.detectorbase.__class__.__name__)

    try:
        print(imgfactory.rawdata.focus())
    except AttributeError:
        # Not all instances have this attribute
        print("  multireadout")

    I_raw_data = imgfactory.get_raw_data()
    if not isinstance(I_raw_data, tuple):
        I_raw_data = (I_raw_data,)

    try:
        R_raw_data = instance.get_raw_data()
    except TypeError:
        R_raw_data = instance.get_raw_data(0)

    if not isinstance(R_raw_data, tuple):
        R_raw_data = (R_raw_data,)

    # NOTE dxtbx and image factory arrays are compared here for identical values.
    for Ip, Rp in zip(I_raw_data, R_raw_data):
        assert (Ip == Rp).all_eq(True)


@pytest.mark.parametrize("test_image", _files)
def test_read_image(test_image):
    """Test that all the regression images can be read"""

    if not h5py and test_image.endswith((".h5", ".nxs")):
        pytest.skip("could not import 'h5py'")

    format_instance = dxtbx.format.Registry.get_format_class_for_file(test_image)
    print("Reading", test_image)
    print("Format:", format_instance)
    assert format_instance, "no matching format class found"
    instance = format_instance(test_image)

    # Test metadata reading
    instance.get_goniometer()
    instance.get_beam()
    instance.get_scan()
    detector = instance.get_detector()
    # From old test_dxtbx; get the image size
    if detector is not None:
        detector[0].get_image_size()

    for panel in detector:
        d_mat = scitbx.matrix.sqr(panel.get_d_matrix())
        if d_mat.determinant() <= 0:
            print("  d matrix with non-positive determinant")

    # Test reading of the raw data
    # XDS, HKL we expect to fail for this  - so skip this part for those
    if not test_image.endswith(("XDS", "HKL")):
        try:
            R_raw_data = instance.get_raw_data()
        except TypeError:
            R_raw_data = instance.get_raw_data(0)

        if not isinstance(R_raw_data, tuple):
            R_raw_data = (R_raw_data,)

        print("%-40s" % format_instance.__name__, R_raw_data[0].focus())

        # Specific test for cctbx/dxtbx#163. This test will fail if char is unsigned.
        if "APS_24IDC" in test_image and "pilatus_1_0001.cbf" in test_image:
            d = R_raw_data[0]
            assert (
                flex.sum(d.as_1d().select(d.as_1d() >= 0)) == 20108255
            )  # matches total counts from dxtbx.print_header


@pytest.mark.parametrize("test_image", _files)
def test_format_class_API_assumptions(test_image):
    """For a given image file, walk the whole DAG of Format objects, and
    verify the following basic assumptions for format classes:
    * Any file must only be understood by a single leaf node format class.
    * No .understand() call on any top level format class or a child class
      of another understanding format is allowed to throw an exception.
    """
    if not h5py and test_image.endswith((".h5", ".nxs")):
        pytest.skip("could not import 'h5py'")

    dag = dxtbx.format.Registry.get_format_class_dag()

    def recurse(parentformat, filename, level=0):
        known_format_class = None
        multiple_formats = False
        for subformat in dag.get(parentformat, []):
            format_class = dxtbx.format.Registry.get_format_class_for(subformat)
            if get_url_scheme(filename) not in format_class.schemes:
                print("Not matching ", filename, "to", format_class)
                continue
            understood = format_class.understand(filename)
            print("%s%s: %s" % ("  " * level, subformat, understood))
            if understood:
                recursive_format_class, subtree_multiple = recurse(
                    subformat, filename, level + 1
                )
                understood_format_class = recursive_format_class or subformat
                if known_format_class and known_format_class != understood_format_class:
                    print(
                        f"File can be understood as {known_format_class} and {understood_format_class}"
                    )
                    multiple_formats = True
                known_format_class = understood_format_class
                multiple_formats |= subtree_multiple
        return known_format_class, multiple_formats

    understood_format, multiple_formats = recurse("Format", test_image)

    assert not multiple_formats, "image file understood by multiple Format objects"
    # It's a failure if nothing could understand this file
    assert understood_format, "No formatter could be found"
    print("File understood as", understood_format)



from __future__ import annotations

import copy

import pytest

from libtbx.phil import parse
from scitbx import matrix

from dxtbx.model import Beam, PolychromaticBeam
from dxtbx.model.beam import BeamFactory, Probe, beam_phil_scope


def test_setting_direction_and_wavelength():
    direction = matrix.col((0.013142, 0.002200, 1.450476))
    unit_direction = direction.normalize()
    wavelength = 0.689400

    # Create the beam
    b = Beam(direction, wavelength)

    eps = 1e-7

    # Check direction is a unit vector
    assert matrix.col(b.get_sample_to_source_direction()).length() == pytest.approx(1)
    assert abs(matrix.col(b.get_sample_to_source_direction()) - unit_direction) <= eps

    # Check wavelength is correct
    assert b.get_wavelength() == pytest.approx(wavelength)

    # Check s0 is in direction and has length 1/wavelength
    assert matrix.col(b.get_s0()).length() == pytest.approx(1.0 / wavelength)
    assert abs(-matrix.col(b.get_s0()).normalize() - unit_direction) <= eps


def test_setting_s0():
    direction = matrix.col((0.013142, 0.002200, 1.450476))
    unit_direction = direction.normalize()
    wavelength = 0.689400
    s0 = -unit_direction * 1.0 / wavelength

    # Create the beam
    b = Beam(s0)

    eps = 1e-7

    # Check direction is a unit vector
    assert matrix.col(b.get_sample_to_source_direction()).length() == pytest.approx(1)
    assert abs(matrix.col(b.get_sample_to_source_direction()) - unit_direction) <= eps

    # Check wavelength is correct
    assert b.get_wavelength() == pytest.approx(wavelength)

    # Check s0 is in direction and has length 1/wavelength
    assert matrix.col(b.get_s0()).length() == pytest.approx(1.0 / wavelength)
    assert abs(-matrix.col(b.get_s0()).normalize() - unit_direction) <= eps
    assert abs(matrix.col(b.get_s0()) - s0) <= eps


def test_from_phil():
    direction = matrix.col((0.013142, 0.002200, 1.450476))
    wavelength = 0.689400

    reference = Beam(direction, wavelength)

    params1 = beam_phil_scope.fetch(
        parse(
            """
    beam {
      wavelength = 1.0
      direction = (0, 0, 1)
    }
  """
        )
    ).extract()

    params2 = beam_phil_scope.fetch(
        parse(
            """
    beam {
      wavelength = 1.0
    }
  """
        )
    ).extract()

    # Create the beam
    assert BeamFactory.from_phil(params1)
    assert BeamFactory.from_phil(params2, reference)
    with pytest.raises(RuntimeError):
        BeamFactory.from_phil(params2)

    params3 = beam_phil_scope.fetch(
        parse(
            """
    beam {
      polarization_normal = 1,0,0
      polarization_fraction = 0.5
    }
  """
        )
    ).extract()
    b3 = BeamFactory.from_phil(params3, reference)
    assert b3.get_polarization_fraction() == 0.5
    assert b3.get_polarization_normal() == (1.0, 0.0, 0.0)

    params3 = beam_phil_scope.fetch(
        parse(
            """
    beam {
        probe = electron
        sample_to_source_distance = 4000
    }
  """
        )
    ).extract()
    b4 = BeamFactory.from_phil(params3, reference)
    assert b4.get_probe() == Probe.electron
    assert b4.get_sample_to_source_distance() == pytest.approx(4000)


def test_scan_varying():
    direction = matrix.col((0.013142, 0.002200, 1.450476))
    unit_direction = direction.normalize()
    wavelength = 0.689400
    s0 = -unit_direction * 1.0 / wavelength

    # Create the beam
    b = Beam(s0)

    assert b.get_num_scan_points() == 0
    assert b.get_s0_at_scan_points().size() == 0
    with pytest.raises(RuntimeError):
        b.get_s0_at_scan_point(0)  # should raise RuntimeError

    # set varying beam
    num_scan_points = 11
    s0_static = matrix.col(b.get_s0())
    s0_as_scan_points = [s0_static]
    axis = matrix.col.random(3, -1.0, 1.0).normalize()
    for i in range(num_scan_points - 1):
        s0_as_scan_points.append(
            s0_as_scan_points[-1].rotate_around_origin(axis, angle=0.01, deg=True)
        )
    b.set_s0_at_scan_points(s0_as_scan_points)
    assert b.get_num_scan_points() == 11
    assert b.get_s0_at_scan_points().size() == 11

    for t in range(num_scan_points):
        s0_t = matrix.col(b.get_s0_at_scan_point(t))
        assert s0_t == s0_as_scan_points[t]

    # also test setting as tuple
    b.set_s0_at_scan_points(tuple(s0_as_scan_points))
    assert b.get_num_scan_points() == 11
    assert b.get_s0_at_scan_points().size() == 11

    # test resetting
    b.reset_scan_points()
    assert b.get_num_scan_points() == 0
    assert b.get_s0_at_scan_points().size() == 0


def test_beam_object_comparison():
    direction = matrix.col((0.013142, 0.002200, 1.450476))
    unit_direction = direction.normalize()
    wavelength = 0.689400
    s0 = -unit_direction * 1.0 / wavelength

    # Equal beams with scan-points set
    b1 = Beam(s0)
    b1.set_s0_at_scan_points([s0] * 5)
    b2 = Beam(s0)
    b2.set_s0_at_scan_points([s0] * 5)

    assert b1 == b2
    assert b1.is_similar_to(b2)

    # Different direction
    b3 = Beam(-s0)
    b3.set_s0_at_scan_points([-s0] * 5)
    assert b1 != b3
    assert not b1.is_similar_to(b3)

    # Different wavelength
    b4 = Beam(s0 * 1.5)
    b4.set_s0_at_scan_points([s0 * 1.5] * 5)
    assert b1 != b4
    assert not b1.is_similar_to(b4)


def test_beam_self_serialization():
    beam = Beam()
    assert beam == BeamFactory.from_dict(beam.to_dict())


def test_polychromatic_beam_from_phil():
    params = beam_phil_scope.fetch(
        parse(
            """
    beam {
      type = polychromatic
      direction = (0., 0., 1.)
      divergence = 0.2
      sigma_divergence = 0.3
      polarization_normal = (0., -1., 0.)
      polarization_fraction = .65
      transmission = .5
      flux = .75
      sample_to_source_distance = 5000
      wavelength_range = (0.2, 10)
    }
    """
        )
    ).extract()

    beam = BeamFactory.from_phil(params)
    assert isinstance(beam, PolychromaticBeam)

    assert beam.get_sample_to_source_direction() == pytest.approx((0.0, 0.0, 1.0))
    assert beam.get_divergence() == pytest.approx(0.2)
    assert beam.get_sigma_divergence() == pytest.approx(0.3)
    assert beam.get_polarization_normal() == pytest.approx((0.0, -1.0, 0.0))
    assert beam.get_polarization_fraction() == pytest.approx(0.65)
    assert beam.get_transmission() == pytest.approx(0.5)
    assert beam.get_flux() == pytest.approx(0.75)
    assert beam.get_sample_to_source_distance() == pytest.approx(5000)
    assert beam.get_wavelength_range() == pytest.approx((0.2, 10))


def test_polychromatic_beam_from_dict():
    beam = PolychromaticBeam()
    assert beam == BeamFactory.from_dict(beam.to_dict())


def test_make_polychromatic_beam():
    direction = (0.0, 0.0, 1.0)
    divergence = 0.2
    sigma_divergence = 0.3
    polarization_normal = (0.0, -1.0, 0.0)
    polarization_fraction = 0.65
    transmission = 0.5
    flux = 0.75
    probe = Probe.neutron
    sample_to_source_distance = 8500
    wavelength_range = (0.2, 10)

    beam = BeamFactory.make_polychromatic_beam(
        direction=direction,
        divergence=divergence,
        sigma_divergence=sigma_divergence,
        polarization_normal=polarization_normal,
        polarization_fraction=polarization_fraction,
        transmission=transmission,
        flux=flux,
        probe=probe,
        sample_to_source_distance=sample_to_source_distance,
        wavelength_range=wavelength_range,
    )

    assert beam.get_sample_to_source_direction() == pytest.approx((0.0, 0.0, 1.0))
    assert beam.get_divergence() == pytest.approx(0.2)
    assert beam.get_sigma_divergence() == pytest.approx(0.3)
    assert beam.get_polarization_normal() == pytest.approx((0.0, -1.0, 0.0))
    assert beam.get_polarization_fraction() == pytest.approx(0.65)
    assert beam.get_transmission() == pytest.approx(0.5)
    assert beam.get_flux() == pytest.approx(0.75)
    assert beam.get_probe() == Probe.neutron
    assert beam.get_sample_to_source_distance() == pytest.approx(8500.0)
    assert beam.get_wavelength_range() == pytest.approx((0.2, 10))


def test_polychromatic_beam_wavelength_guards():
    beam = PolychromaticBeam()
    with pytest.raises(RuntimeError):
        _ = beam.get_wavelength()
    with pytest.raises(RuntimeError):
        _ = beam.get_s0()
    with pytest.raises(RuntimeError):
        _ = beam.get_num_scan_points()
    with pytest.raises(RuntimeError):
        _ = beam.get_s0_at_scan_points()
    with pytest.raises(RuntimeError):
        _ = beam.get_s0_at_scan_point(0)
    with pytest.raises(RuntimeError):
        beam.reset_scan_points()
    with pytest.raises(RuntimeError):
        beam.set_wavelength(1.0)
    with pytest.raises(RuntimeError):
        beam.set_s0((0.0, 0.0, 0.1))


def test_polychromatic_beam_comparison():
    beam1 = PolychromaticBeam()
    beam1.set_wavelength_range((0.0, 1.0))
    beam1.set_sample_to_source_distance(10)
    beam1.set_direction((0.0, 0.0, 1.0))
    beam2 = PolychromaticBeam()
    beam2.set_wavelength_range((0.0, 1.0))
    beam2.set_sample_to_source_distance(10)
    beam2.set_direction((0.0, 0.0, 1.0))

    assert beam1 == beam2
    beam1.set_wavelength_range((0.0, 2.0))
    assert beam1 != beam2
    beam1.set_wavelength_range((0.0, 1.0))
    assert beam1 == beam2

    beam1.set_sample_to_source_distance(1.0)
    assert beam1 != beam2
    beam1.set_sample_to_source_distance(10)
    assert beam1 == beam2

    beam1.set_probe(Probe.neutron)
    assert beam1 != beam2
    beam1.set_probe(Probe.xray)
    assert beam1 == beam2

    beam1.set_direction((1.0, 0.0, 0.0))
    assert beam1 != beam2
    beam1.set_direction((0.0, 0.0, 1.0))
    assert beam1 == beam2

    beam3 = Beam()
    assert beam1 != beam3


def test_polychromatic_beam_str():
    beam = PolychromaticBeam()
    assert (
        beam.__str__()
        == "Beam:\n    probe: x-ray\n    sample to source direction : {0,0,1}\n    divergence: 0\n    sigma divergence: 0\n    polarization normal: {0,1,0}\n    polarization fraction: 0.5\n    flux: 0\n    transmission: 1\n    sample to source distance : 0\n    wavelength range : {0,0}\n"
    )


def test_copy_beam():
    beam = PolychromaticBeam()
    assert beam == copy.deepcopy(beam)
    beam = Beam()
    assert beam == copy.deepcopy(beam)



from __future__ import annotations

import pickle

import pytest

import cctbx.array_family.flex as flex

from dxtbx.model import Scan, ScanFactory


def test_scan_to_string_does_not_crash_on_empty_scan():
    print(Scan())


def test_scan_wrap_around_zero():
    filenames = [f"foobar_{n:03d}.cbf" for n in range(350, 370)]
    starts = [r % 360 for r in range(350, 370)]
    ends = [(r + 1) % 360 for r in range(350, 370)]
    scans = [
        ScanFactory.single_file(f, [0.1], s, e - s, 0)
        for f, s, e in zip(filenames, starts, ends)
    ]
    s0 = scans[0]
    for s in scans[1:]:
        assert s.get_oscillation()[1] == 1
        s0 += s
    assert s0.get_oscillation() == pytest.approx((350, 1))
    assert s0.get_image_range() == (350, 369)


def test_make_scan_from_properties():
    image_range = (1, 10)

    properties = {}
    scan = ScanFactory.make_scan_from_properties(
        image_range=image_range, properties=properties
    )
    assert scan.get_properties() == {}

    properties = {
        "test_int": tuple(range(10)),
        "test_float": tuple([float(i * 2) for i in range(10)]),
        "test_bool": tuple([True for i in range(10)]),
        "test_string": tuple([f"test_{i}" for i in range(10)]),
        "test_vec3_double": tuple([(1.0, 1.0, 1.0) for i in range(10)]),
        "test_vec2_double": tuple([(2.0, 2.0) for i in range(10)]),
    }

    scan = ScanFactory.make_scan_from_properties(
        image_range=image_range, properties=properties
    )

    result = scan.get_properties()
    for key, value in properties.items():
        assert key in result
        assert len(value) == len(result[key])
        for i in range(len(value)):
            assert result[key][i] == pytest.approx(value[i])


def test_set_get_properties():
    image_range = (1, 10)
    scan = ScanFactory.make_scan_from_properties(image_range=image_range, properties={})

    test_int = flex.int(10, 1)
    test_float = flex.double(10, 1.0)
    test_bool = flex.bool(10, True)
    test_string = flex.std_string(10, "Test")
    test_vec3_double = flex.vec3_double(10, (1.0, 1.0, 1.0))
    test_vec2_double = flex.vec2_double(10, (2.0, 2.0))

    scan.set_property("test_int", test_int)
    result = scan.get_property("test_int")
    for idx, i in enumerate(test_int):
        assert result[idx] == pytest.approx(i)

    scan.set_property("test_float", test_float)
    result = scan.get_property("test_float")
    for idx, i in enumerate(test_float):
        assert result[idx] == pytest.approx(i)

    scan.set_property("test_bool", test_bool)
    result = scan.get_property("test_bool")
    for idx, i in enumerate(test_bool):
        assert result[idx] == pytest.approx(i)

    scan.set_property("test_string", test_string)
    result = scan.get_property("test_string")
    for idx, i in enumerate(test_string):
        assert result[idx] == pytest.approx(i)

    scan.set_property("test_vec3_double", test_vec3_double)
    result = scan.get_property("test_vec3_double")
    for idx, i in enumerate(test_vec3_double):
        assert result[idx] == pytest.approx(i)

    scan.set_property("test_vec2_double", test_vec2_double)
    result = scan.get_property("test_vec2_double")
    for idx, i in enumerate(test_vec2_double):
        assert result[idx] == pytest.approx(i)

    # Test incompatible size with property table

    test_int = flex.int(5, 1)
    with pytest.raises(RuntimeError):
        scan.set_property("test_int", test_int)

    test_float = flex.double(5, 1.0)
    with pytest.raises(RuntimeError):
        scan.set_property("test_float", test_float)

    test_bool = flex.bool(5, True)
    with pytest.raises(RuntimeError):
        scan.set_property("test_bool", test_bool)

    test_string = flex.std_string(5, "Test")
    with pytest.raises(RuntimeError):
        scan.set_property("test_string", test_string)

    test_vec3_double = flex.vec3_double(5, (1.0, 1.0, 1.0))
    with pytest.raises(RuntimeError):
        scan.set_property("test_vec3_double", test_vec3_double)

    test_vec2_double = flex.vec2_double(5, (2.0, 2.0))
    with pytest.raises(RuntimeError):
        scan.set_property("test_vec2_double", test_vec2_double)

    test_int = flex.int(15, 1)
    with pytest.raises(RuntimeError):
        scan.set_property("test_int", test_int)

    test_float = flex.double(15, 1.0)
    with pytest.raises(RuntimeError):
        scan.set_property("test_float", test_float)

    test_bool = flex.bool(15, True)
    with pytest.raises(RuntimeError):
        scan.set_property("test_bool", test_bool)

    test_string = flex.std_string(15, "Test")
    with pytest.raises(RuntimeError):
        scan.set_property("test_string", test_string)

    test_vec3_double = flex.vec3_double(15, (1.0, 1.0, 1.0))
    with pytest.raises(RuntimeError):
        scan.set_property("test_vec3_double", test_vec3_double)

    test_vec2_double = flex.vec2_double(15, (2.0, 2.0))
    with pytest.raises(RuntimeError):
        scan.set_property("test_vec2_double", test_vec2_double)


def test_scan_properties_pickle():
    image_range = (1, 10)
    properties = {
        "test_int": tuple(range(10)),
        "test_float": tuple([float(i * 2) for i in range(10)]),
        "test_bool": tuple([True for i in range(10)]),
        "test_string": tuple([f"test_{i}" for i in range(10)]),
        "test_vec3_double": tuple([(1.0, 1.0, 1.0) for i in range(10)]),
        "test_vec2_double": tuple([(2.0, 2.0) for i in range(10)]),
    }

    scan = ScanFactory.make_scan_from_properties(
        image_range=image_range, properties=properties
    )

    obj = pickle.dumps(scan)
    scan2 = pickle.loads(obj)
    assert scan == scan2

    result = scan.get_properties()
    for key, value in properties.items():
        assert key in result
        assert len(value) == len(result[key])
        for i in range(len(value)):
            assert result[key][i] == pytest.approx(value[i])


def test_scan_properties_to_dict():
    image_range = (1, 10)
    properties = {
        "test_int": tuple(range(10)),
        "test_float": tuple([float(i * 2) for i in range(10)]),
        "test_bool": tuple([True for i in range(10)]),
        "test_string": tuple([f"test_{i}" for i in range(10)]),
        "test_vec3_double": tuple([(1.0, 1.0, 1.0) for i in range(10)]),
        "test_vec2_double": tuple([(2.0, 2.0) for i in range(10)]),
    }

    scan = ScanFactory.make_scan_from_properties(
        image_range=image_range, properties=properties
    )

    expected_result = {
        "test_bool": (True, True, True, True, True, True, True, True, True, True),
        "test_float": (0.0, 2.0, 4.0, 6.0, 8.0, 10.0, 12.0, 14.0, 16.0, 18.0),
        "test_int": (0, 1, 2, 3, 4, 5, 6, 7, 8, 9),
        "test_string": (
            "test_0",
            "test_1",
            "test_2",
            "test_3",
            "test_4",
            "test_5",
            "test_6",
            "test_7",
            "test_8",
            "test_9",
        ),
        "test_vec2_double": (
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
            (2.0, 2.0),
        ),
        "test_vec3_double": (
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
            (1.0, 1.0, 1.0),
        ),
    }

    result = scan.to_dict()["properties"]
    for key, value in expected_result.items():
        assert key in result
        assert len(value) == len(result[key])
        for i in range(len(value)):
            assert result[key][i] == pytest.approx(value[i])


def test_scan_properties_equivalence():
    int_diff = 1
    double_diff = 1e-6

    image_range = (1, 10)
    s1 = ScanFactory.make_scan_from_properties(image_range=image_range, properties={})
    s2 = ScanFactory.make_scan_from_properties(image_range=image_range, properties={})

    # Empty table

    assert s1 == s2

    # int columns

    test_int = flex.int(10, 1)
    test_int2 = flex.int(10, 1)
    test_int2[0] += int_diff

    s1.set_property("test_int", test_int)

    assert s1 != s2

    s2.set_property("test_int", test_int2)

    assert s1 != s2

    test_int2[0] -= int_diff

    s2.set_property("test_int", test_int2)

    assert s1 == s2

    # double columns

    test_double = flex.double(10, 1.0)
    test_double2 = flex.double(10, 1.0)
    test_double2[0] += double_diff

    s1.set_property("test_double", test_double)

    assert s1 != s2

    s2.set_property("test_double", test_double2)

    assert s1 != s2

    test_double2[0] -= double_diff

    s2.set_property("test_double", test_double2)

    assert s1 == s2

    # bool columns

    test_bool = flex.bool(10, True)
    test_bool2 = flex.bool(10, True)
    test_bool2[0] = False

    s1.set_property("test_bool", test_bool)

    assert s1 != s2

    s2.set_property("test_bool", test_bool2)

    assert s1 != s2

    test_bool2[0] = True

    s2.set_property("test_bool", test_bool2)

    assert s1 == s2

    # string columns

    test_string = flex.std_string(10, "Test")
    test_string2 = flex.std_string(10, "Test")
    test_string2[0] = "Test2"

    s1.set_property("test_string", test_string)

    assert s1 != s2

    s2.set_property("test_string", test_string2)

    assert s1 != s2

    test_string2[0] = "Test"

    s2.set_property("test_string", test_string2)

    assert s1 == s2

    #  vec3<double> columns

    test_vec3_double = flex.vec3_double(10, (1.0, 1.0, 1.0))
    test_vec3_double2 = flex.vec3_double(10, (1.0, 1.0, 1.0))
    test_vec3_double2[0] = (1.0 + double_diff, 1.0 + double_diff, 1.0 + double_diff)

    s1.set_property("test_vec3_double", test_vec3_double)

    assert s1 != s2

    s2.set_property("test_vec3_double", test_vec3_double2)

    assert s1 != s2
    vec = test_vec3_double2[0]
    test_vec3_double2[0] = (
        vec[0] - double_diff,
        vec[1] - double_diff,
        vec[2] - double_diff,
    )

    s2.set_property("test_vec3_double", test_vec3_double2)

    assert s1 == s2

    #  vec2<double> columns

    test_vec2_double = flex.vec2_double(10, (2.0, 2.0))
    test_vec2_double2 = flex.vec2_double(10, (2.0, 2.0))
    test_vec2_double2[0] = (2.0 + double_diff, 2.0 + double_diff)

    s1.set_property("test_vec2_double", test_vec2_double)

    assert s1 != s2

    s2.set_property("test_vec2_double", test_vec2_double2)

    assert s1 != s2

    vec = test_vec2_double2[0]
    test_vec2_double2[0] = (vec[0] - double_diff, vec[1] - double_diff)

    s2.set_property("test_vec2_double", test_vec2_double2)

    assert s1 == s2


def test_scan_constant_oscillation_width():
    image_range = (1, 10)
    scan = ScanFactory.make_scan_from_properties(image_range=image_range, properties={})

    oscillation = flex.double(10, 1.0)
    scan.set_property("oscillation", oscillation)
    oscillation[3] = 2.0
    with pytest.raises(RuntimeError):
        scan.set_property("oscillation", oscillation)

    oscillation = flex.double(10, 1.0)
    properties = {"oscillation": oscillation}
    scan.set_properties(properties)
    properties["oscillation"][3] = 2.0
    with pytest.raises(RuntimeError):
        scan.set_properties(properties)

    oscillation = flex.double(10, 1.0)
    properties = {"oscillation": oscillation}
    scan = ScanFactory.make_scan_from_properties(
        image_range=image_range, properties=properties
    )
    properties["oscillation"][3] = 2.0
    with pytest.raises(RuntimeError):
        scan = ScanFactory.make_scan_from_properties(
            image_range=image_range, properties=properties
        )


def test_print_scan():
    image_range = (1, 10)
    scan = ScanFactory.make_scan_from_properties(image_range=image_range, properties={})

    expected_scan_string = (
        "Scan:\n    number of images:   10\n    image range:   {1,10}\n"
    )
    assert scan.__str__() == expected_scan_string

    # Properties with unique output

    scan.set_oscillation((0.0, 0.5))
    expected_scan_string = "Scan:\n    number of images:   10\n    image range:   {1,10}\n    oscillation:   {0,0.5}\n"
    assert scan.__str__() == expected_scan_string

    scan.set_property("epochs", flex.double(10))
    expected_scan_string = "Scan:\n    number of images:   10\n    image range:   {1,10}\n    epoch:    0\n    oscillation:   {0,0.5}\n"
    assert scan.__str__() == expected_scan_string

    scan.set_property("exposure_time", flex.double(10))
    expected_scan_string = "Scan:\n    number of images:   10\n    image range:   {1,10}\n    epoch:    0\n    exposure time:    0\n    oscillation:   {0,0.5}\n"
    assert scan.__str__() == expected_scan_string

    # Generic properties

    properties = {
        "test_int": tuple(range(10)),
        "test_float": tuple([float(i * 2) for i in range(10)]),
        "test_bool": tuple([True for i in range(10)]),
        "test_string": tuple([f"test_{i}" for i in range(10)]),
        "test_vec3_double": tuple([(1.0, 1.0, 1.0) for i in range(10)]),
        "test_vec2_double": tuple([(2.0, 2.0) for i in range(10)]),
    }

    scan = ScanFactory.make_scan_from_properties(
        image_range=image_range, properties=properties
    )
    expected_scan_string = "Scan:\n    number of images:   10\n    image range:   {1,10}\n    test_bool:    1 - 1\n    test_float:    0 - 18\n    test_int:    0 - 9\n    test_string:    test_0 - test_9\n    test_vec2_double:    {2,2} - {2,2}\n    test_vec3_double:    {1,1,1} - {1,1,1}\n"
    assert scan.__str__() == expected_scan_string


def test_scan_is_still():
    scan = Scan()
    assert scan.is_still()
    scan = Scan((1, 1000), (0, 0.05))
    assert not scan.is_still()
    # Specifically test the bug from dxtbx#4 triggered by ending on 0°
    scan = Scan((1, 1800), (-90, 0.05))
    assert not scan.is_still()
    scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"time_of_flight": list(range(10))}
    )
    assert not scan.is_still()
    scan = ScanFactory.make_scan_from_properties(
        (1, 10), properties={"other_property": list(range(10))}
    )
    assert scan.is_still()


def test_scan_properties_from_dict():
    image_range = (1, 10)
    properties = {"test": list(range(10))}
    scan = ScanFactory.make_scan_from_properties(image_range, properties)
    assert scan == ScanFactory.from_dict(scan.to_dict())

    image_range = (1, 1)
    properties = {"oscillation": [1.0], "oscillation_width": [0.5]}
    scan = ScanFactory.make_scan_from_properties(image_range, properties)
    assert scan == ScanFactory.from_dict(scan.to_dict())

    scan = ScanFactory.from_dict(
        {
            "oscillation": [1.0, 0.5],
            "image_range": [1, 1],
            "exposure_time": [0.5],
            "epochs": [1],
        }
    )
    assert scan == ScanFactory.from_dict(scan.to_dict())



from __future__ import annotations

try:
    from importlib import metadata

    version = metadata.version("dxtbx")
except ImportError:
    try:
        # Running on pre-3.8 Python; use importlib-metadata backport package
        import importlib_metadata as metadata  # type: ignore

        version = metadata.version("dxtbx")
    except ImportError:
        # If we don't have importlib or the backport assume 3.8 series.
        # In this scenario we probably aren't going to be checking this anyway.
        version = "3.8.dev"



from __future__ import annotations

from boost_adaptbx.boost.python import streambuf
from scitbx.array_family import flex

def compress(array: flex.int) -> bytes: ...
def is_big_endian() -> bool: ...
def read_float32(file: streambuf, count: int) -> float: ...
def read_int16(file: streambuf, count: int) -> int: ...
def read_int32(file: streambuf, count: int) -> int: ...
def read_uint16(file: streambuf, count: int) -> int: ...
def read_uint16_bs(file: streambuf, count: int) -> int: ...
def read_uint32(file: streambuf, count: int) -> int: ...
def read_uint32_bs(file: streambuf, count: int) -> int: ...
def read_uint8(file: streambuf, count: int) -> int: ...
def uncompress(packed: bytes, slow: int, fast: int) -> flex.int: ...
